# 2010 July 9 by Joshua R. Davis

import time
import copy
from utility import *

class Delayer(object):
	"""A function (or method) call that is scheduled to take place sometime in the future."""
	
	def __init__(self, endTime, function, arguments):
		"""Initializes the task and adds it to the current window. At the specified endTime, the window will call the function on the arguments given in the sequence arguments."""
		self.endTime = endTime
		self.function = function
		self.arguments = copy.deepcopy(arguments)
	
	def update(self, newTime):
		"""Non-user. Notifies the task of the current time. If the scheduled time for the task has been reached, then the task is performed and True is returned. Otherwise nothing happens and False is returned."""
		if newTime >= self.endTime:
			self.function(*self.arguments)
			return True
		else:
			return False

class Positioner(object):
	"""Animates a function (or method) call over a specified time period, based on starting and ending values for the arguments to the function. The canonical example is animating object.setPosition() by giving starting and ending positions."""
	
	def __init__(self, startTime, endTime, function, startArgs, endArgs, repetitions=1, rescaler=None):
		"""Initializes the animation to run from startTime to endTime (in seconds, as given by time.time()). The animation behaves as if at startTime function is called with startArgs and at endTime function is called with endArgs. If repetitions is None, then the animation loops forever; otherwise, it loops for the specified whole number of repetitions. The rescaler is an optional function that takes in a float between 0 and 1 and returns a float between 0 and 1. For example, the utility function bopagopa.cubicBijection() causes the animation to begin and finish softly."""
		self.startTime = startTime
		self.endTime = endTime
		self.function = function
		self.startArgs = copy.deepcopy(startArgs)
		self.endArgs = copy.deepcopy(endArgs)
		self.repetitions = repetitions
		self.rescaler = rescaler
	
	def getStartArguments(self):
		"""Returns the starting arguments."""
		return self.startArgs
	
	def getEndArguments(self):
		"""Returns the ending arguments."""
		return self.endArgs
	
	def getRepetitions(self):
		"""Returns the number of repetitions to be performed, which may be a float."""
		return self.repetitions
	
	def internalTime(self, newTime):
		"""Private. Takes as input a time, usually the present time. Returns a pair consisting of an integer and a float. The integer is the number of periods completed at the given time; it may be negative. The float is the fraction (between 0 and 1) of the way through the next period that the given time represents; it has already been rescaled, if a rescaler is in use."""
		quotient = (newTime - self.startTime) / (self.endTime - self.startTime)
		whole = math.floor(quotient)
		fraction = quotient - whole
		if self.rescaler != None:
			fraction = self.rescaler(fraction)
		return (int(whole), fraction)
	
	def update(self, newTime):
		"""Non-user. The input newTime input is a float expressing the time in seconds since the computer's datum, usually obtained from time.time(). If this time is before the start of the animation, then no work is done; if the time is after the start of the animation, then the animation calls its function with its interpolated arguments. The output is a Boolean expressing whether the animation is finished."""
		(numReps, fraction) = self.internalTime(newTime)
		if numReps < 0:
			# The animation hasn't begun yet.
			return False
		elif self.repetitions != None and numReps >= self.repetitions:
			# End the animation.
			self.function(*self.endArgs)
			return True
		else:
			# Continue the animation.
			args = nestInterpolation(self.startArgs, fraction, self.endArgs)
			self.function(*args)
			return False

class Displacer(Positioner):
	"""Animates a function (or method) call over a specified time period, based on a zero value and displacement value for the arguments to the function. The canonical example is animating object.translate() by giving the zero vector and a displacement vector, but can be used in various other situations."""
	
	def __init__(self, startTime, endTime, function, zeroArgs, displacementArgs, repetitions=1, rescaler=None):
		"""Initializes the animation to run from startTime to endTime (in seconds, as given by time.time()). The animation behaves as if at startTime function is called with zeroArgs, and upon each time step delta function is called with zeroArgs + delta * displacementArgs. If repetitions is None, then the animation loops forever; otherwise, it loops for the specified whole number of repetitions. The rescaler is an optional function that takes in a float between 0 and 1 and returns a float between 0 and 1. For example, the utility function bopagopa.cubicBijection() causes the animation to begin and finish softly."""
		Positioner.__init__(self, startTime, endTime, function, zeroArgs, displacementArgs, repetitions, rescaler)
		self.numReps = 0
		self.fraction = 0.0
	
	def update(self, newTime):
		"""Non-user. Overrides Positioner's version to update based on the change in time, rather than the time itself."""
		(numReps, fraction) = self.internalTime(newTime)
		if numReps < 0:
			# The animation hasn't begun yet.
			return False
		else:
			# Determine whether the animation is finished.
			repetitions = self.getRepetitions()
			if repetitions != None and numReps >= repetitions:
				numReps = repetitions
				fraction = 0.0
				finished = True
			else:
				finished = False
			# Perform the next frame of the animation.
			delta = (numReps - self.numReps) + (fraction - self.fraction)
			self.numReps = numReps
			self.fraction = fraction
			args = nestCombination(self.getStartArguments(), delta, self.getEndArguments())
			self.function(*args)
			return finished

