# 2010 October 5 by Joshua R. Davis

import OpenGL.GL as gl
import ode
import math
import window
from utility import *

class Body3D(object):
	"""A rigid body in a 3D scene. The visual appearance of the body is determined by a Mesh, a set of Textures, and a set of uniform properties. The dynamics of the body are handled by an underlying Open Dynamics Engine Body object. All messages that the Body3D receives and doesn't recognize are delegated to the ODE Body. In most cases this means that you can talk to Body3D as if it's a subclass of ODE's Body. For example, you can set the Body3D's position by calling myBody3D.setPosition(xyz) instead of myBody3D.getBody().setPosition(xyz)."""
	
	def __init__(self, mesh=None, textures=[], uniforms=[], world=None):
		"""Initializes the body with the given Mesh, list of Textures, and list of uniforms. These data must be compatible with whichever renderer the body is eventually attached to. The world argument should be an ODE World object, where the body will live; if world is None, then the body lives in the world of the current window."""
		if world == None:
			world = window.getWindow().getWorld()
		# This is like self.body = ode.Body(world), but doesn't trigger __setattr__.
		self.__dict__["body"] = ode.Body(world)
		self.uniforms = uniforms[:]
		self.textures = textures[:]
		self.setMesh(mesh)
		# !! set mass, collision properties, other dynamics?
	
	def __getattr__(self, attr):
		"""Non-user method; part of the system for delegating unrecognized messages to the underlying ODE Body."""
		return getattr(self.body, attr)
	
	def __setattr__(self, attr, value):
		"""Non-user method; part of the system for delegating unrecognized messages to the underlying ODE Body."""
		return setattr(self.body, attr, value)
	
	def getExampleBody(self):
		"""Non-user method used by Renderer3D. If this body is ready to draw, as indicated by having a mesh, then returns self. Otherwise returns None."""
		if self.mesh == None:
			return None
		else:
			return self
	
	def getBody(self):
		"""Returns the underlying ODE Body, which is never None. Note however that in most cases you do not need to get the ODE Body; all messages that the Body3D receives, that it doesn't recognize, are delegated to the ODE Body automatically."""
		return self.body
	
	def setMesh(self, mesh):
		"""Sets the mesh. The user is responsible for its being compatible with the renderer."""
		self.mesh = mesh
	
	def getMesh(self):
		"""Returns the mesh."""
		return self.mesh
	
	def getTextures(self):
		"""Returns the list of textures, which the user can alter (append, pop, etc.) as desired, as long as they remain compatible with the renderer."""
		return self.textures
	
	def getUniforms(self):
		"""Returns the list of uniform properties, which the user can alter (append, pop, etc.) as desired, as long as they remain compatible with the renderer."""
		return self.uniforms
	
	def translate(self, v):
		"""Changes the position by adding v = (dx, dy, dz)."""
		self.body.setPosition(vectorSum(self.body.getPosition(), v))
	
	def rotate(self, angle, axis):
		"""Prepends the given angle-axis rotation to the receiver's orientation."""
		self.body.setQuaternion(quaternionProduct(quaternionFromAngleAxis((angle, axis)), self.body.getQuaternion()))
	
	def revolve(self, angle, axis, center):
		"""Revolves the body about the given center point by the angle-axis rotation; changes position but not orientation."""
		delta = vectorDifference(self.body.getPosition(), center)
		delta = vectorRotated((angle, axis), delta)
		self.body.setPosition(vectorSum(center, delta))
	
	def revolveRotate(self, angle, axis, center):
		"""Revolves the body about the given center point by the given angle-axis rotation, also prepending the rotation to the receiver's orientation, so that the same side of the member always faces the center point."""
		self.revolve(angle, axis, center)
		self.rotate(angle, axis)
	
	def getMatrix(self):
		"""Non-user. Returns the product of the receiver's translation and rotation matrices. This matrix takes local object coordinates to world coordinates."""
		rot = self.body.getRotation()
		pos = self.body.getPosition()
		return (rot[0 * 3 + 0], rot[0 * 3 + 1], rot[0 * 3 + 2], pos[0], rot[1 * 3 + 0], rot[1 * 3 + 1], rot[1 * 3 + 2], pos[1], rot[2 * 3 + 0], rot[2 * 3 + 1], rot[2 * 3 + 2], pos[2], 0.0, 0.0, 0.0, 1.0)
	
	def draw(self, uniformLocations, attributeLocations, matrixLocation, m=None):
		"""Non-user. Draws the body using its renderer's locations. The uniform locations correspond one-to-one to the body's uniforms and the attribute locations correspond one-to-one to the mesh's attributes. Any location that is -1 is ignored. matrixLocation is the location of the 4x4 modeling matrix; m is the modeling matrix established by an enclosing group, or None. Matrices are stored as length-16 vectors in row-major order."""
		if self.mesh != None:
			# Establish the modeling transformation.
			if m == None:
				gl.glUniformMatrix4fv(matrixLocation, 1, gl.GL_TRUE, self.getMatrix())
			else:
				gl.glUniformMatrix4fv(matrixLocation, 1, gl.GL_TRUE, affineMatrixProduct(m, self.getMatrix()))
			# Set the textures and uniforms, and draw the mesh.
			for i in range(len(self.textures)):
				gl.glActiveTexture(gl.GL_TEXTURE0 + i)
				self.textures[i].bind()
			for i in range(len(uniformLocations)):
				if uniformLocations[i] != -1:
					if len(self.uniforms[i]) == 4:
						gl.glUniform4f(uniformLocations[i], self.uniforms[i][0], self.uniforms[i][1], self.uniforms[i][2], self.uniforms[i][3])
					elif len(self.uniforms[i]) == 3:
						gl.glUniform3f(uniformLocations[i], self.uniforms[i][0], self.uniforms[i][1], self.uniforms[i][2])
					elif len(self.uniforms[i]) == 2:
						gl.glUniform2f(uniformLocations[i], self.uniforms[i][0], self.uniforms[i][1])
					else:
						gl.glUniform1f(uniformLocations[i], self.uniforms[i][0])
			self.mesh.draw(attributeLocations)

class Group3D(Body3D):
	"""A body that possesses child bodies. The parent and the children act as if they're glued together; the positions and orientations of the children are relative to the parent. Nevertheless, each child is drawn with their its own color, texture, etc."""
	
	def __init__(self, bodies=[]):
		"""Initializes the group with the given list of bodies."""
		Body3D.__init__(self)
		self.bodies = bodies[:]
	
	def getExampleBody(self):
		"""Non-user method used by Renderer3D. Overrides to return an exemplary body from a child, if there isn't one in the receiver."""
		example = Body3D.getExampleBody(self)
		if example == None:
			for body in self.bodies:
				example = body.getExampleBody()
				if example != None:
					return example
			return None
		else:
			return example
	
	def getBodies(self):
		"""Returns the list of bodies in the group, which the user can alter (append, pop, etc.) as desired, as long as they remain compatible with the renderer."""
		return self.bodies
	
	def draw(self, uniformLocations, attributeLocations, matrixLocation, m=None):
		"""Non-user. Overrides Body3D's version to draw the group's bodies relative to the group."""
		if m == None:
			matrix = self.getMatrix()
		else:
			matrix = affineMatrixProduct(m, self.getMatrix())
		for body in self.bodies:
			if body != None:
				body.draw(uniformLocations, attributeLocations, matrixLocation, matrix)
