# 2010 August 30 by Joshua R. Davis

from bopagopa import *

class WindowJr(Window):
	"""A window object corresponds to a window on the screen in which drawing takes place. The program begins by creating the window."""
	
	def __init__(self, width=512, height=512, title="Bopagopa Jr."):
		"""Initializes the window with some default user interface handlers."""
		Window.__init__(self, width, height, title)
		# Initialize the default user interface handlers.
		self.setKeyboardHandler(self.defaultKeyboardHandler)
		self.setDragHandler2D(self.defaultDragHandler2D)
		self.setDragHandler3D(self.defaultDragHandler3D)
		self.setClickHandler3D(self.defaultClickHandler3D)
		self.setBeginDragHandler3D(self.defaultBeginDragHandler3D)
		self.setDragHandler3D(self.defaultDragHandler3D)
		self.focus = (0.0, 0.0, 0.0)
		# Make a white texture to act as a placeholder where no texture is desired.
		self.whiteTexture = makeWhiteTexture()
		# Attach a 2D renderer that draws shapes with a texture multiplied by a uniform RGBA color.
		renderer = Renderer2D()
		renderer.setTextureCoordinateCodes(["a1"])
		renderer.setColorCode("s0 * u0")
		self.setRenderer2D(renderer)
		# Attach a 3D renderer that draws solids with a texture multiplied by a uniform RGB color.
		renderer = Renderer3D()
		renderer.setTextureCoordinateCodes(["a1"])
		renderer.setDiffuseCode("s0 * u0")
		renderer.setNormalCode("normalize(v0)")
		renderer.setVaryingCodesSizes(["normalize((modelingMatrix * vec4(a2, 0.0)).xyz)"], [3])
		#renderer.setIsSpecular(True)
		self.setRenderer3D(renderer)
		# Attach a perspective camera.
		renderer.setCamera(Camera())
	
	def getBodies2D(self):
		"""Returns the window's list of 2D bodies, which the user can alter (append, pop, etc.) as desired."""
		return self.getRenderer2D().getBodies()
	
	def appendBody2D(self, body):
		"""Shortcut for appending a body to the list of 2D bodies."""
		self.getRenderer2D().getBodies().append(body)
	
	def getBodies3D(self):
		"""Returns the window's list of 3D bodies, which the user can alter (append, pop, etc.) as desired."""
		return self.getRenderer3D().getBodies()
	
	def appendBody3D(self, body):
		"""Shortcut for appending a body to the list of 2D bodies."""
		self.getRenderer3D().getBodies().append(body)
	
	def getLights(self):
		"""Returns the window's lights, as a tuple."""
		return self.getRenderer3D().getLights()
	
	def setLights(self, lights):
		"""Sets the window's lights to the given tuple."""
		return self.getRenderer3D().setLights(lights)
	
	def getCamera(self):
		"""Returns the window's camera."""
		return self.getRenderer3D().getCamera()
	
	def setAmbientColor(self, rgb):
		"""Sets the ambient light level."""
		self.getRenderer3D().setAmbientColor(rgb)
	
	def getWhiteTexture(self):
		"""Non-user method used by shapes with no texture, to get a texture that is visually equivalent to no texture."""
		return self.whiteTexture
	
	def defaultKeyboardHandler(self, xy, key):
		"""Provides some basic viewing controls."""
		camera = self.getRenderer3D().getCamera()
		if key == 'w':
			self.setIsWireframe(not self.getIsWireframe())
			self.needsRedrawing()
		elif key == 'f':
			self.setIsFullScreen(not self.getIsFullScreen())
		elif key == '4':
			#camera.revolveRotate(-math.pi / 6.0, (0.0, 0.0, 1.0), self.focus)
			#self.needsRedrawing()
			self.makeDisplacer(0.1, camera.revolveRotate, (0.0, (0.0, 0.0, 1.0), self.focus), (-math.pi / 6.0, (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)), 1, cubicBijection)
		elif key == '6':
			#camera.revolveRotate(math.pi / 6.0, (0.0, 0.0, 1.0), self.focus)
			#self.needsRedrawing()
			self.makeDisplacer(0.1, camera.revolveRotate, (0.0, (0.0, 0.0, 1.0), self.focus), (math.pi / 6.0, (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)), 1, cubicBijection)
		elif key == '2':
			u = camera.getBody().vectorToWorld((1.0, 0.0, 0.0))
			#camera.revolveRotate(math.pi / 12.0, u, self.focus)
			#self.needsRedrawing()
			self.makeDisplacer(0.1, camera.revolveRotate, (0.0, u, self.focus), (math.pi / 12.0, (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)), 1, cubicBijection)
		elif key == '8':
			u = camera.getBody().vectorToWorld((1.0, 0.0, 0.0))
			#camera.revolveRotate(-math.pi / 12.0, u, self.focus)
			#self.needsRedrawing()
			self.makeDisplacer(0.1, camera.revolveRotate, (0.0, u, self.focus), (-math.pi / 12.0, (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)), 1, cubicBijection)
		elif key == '1':
			#camera.dolly(-1.0)
			#self.needsRedrawing()
			self.makeDisplacer(0.1, camera.dolly, (0.0,), (-1.0,), 1, cubicBijection)
		elif key == '7':
			#camera.dolly(1.0)
			#self.needsRedrawing()
			self.makeDisplacer(0.1, camera.dolly, (0.0,), (1.0,), 1, cubicBijection)
		elif key == '5':
			# bug here !!!!
			self.getLights()[0].containPoints([(0.0, 0.0, 0.0), (10.0, 0.0, 0.0), (10.0, 10.0, 0.0), (0.0, 10.0, 0.0), (0.0, 0.0, 4.0), (10.0, 0.0, 4.0), (10.0, 10.0, 4.0), (0.0, 10.0, 4.0)])
			self.needsRedrawing()
	
	def defaultDragHandler2D(self, xyStarting, xyPrevious, xyCurrent, name):
		"""Lets the user drag the 2D bodies around."""
		self.getRenderer2D().getBodies()[name].translate((xyCurrent[0] - xyPrevious[0], xyCurrent[1] - xyPrevious[1]))
		self.needsRedrawing()
	
	def defaultClickHandler3D(self, xy, name, xyz):
		"""Translates the focus of the camera."""
		delta = vectorDifference(xyz, self.focus)
		self.focus = xyz
		body = self.getRenderer3D().getCamera()
		self.makeDisplacer(0.2, body.translate, ((0.0, 0.0, 0.0),), (delta,), 1, cubicBijection)
	
	def defaultBeginDragHandler3D(self, xyStarting, xyCurrent, name, xyzStarting):
		"""Prepares for defaultDragHandler3D()."""
		center = self.getRenderer3D().getBodies()[name].getBody().getPosition()
		# Find the sphere centered on the picked body passing through xyzStarting.
		radial = vectorDifference(xyzStarting, center)
		self.dragRadiusSquared = dotProduct(radial, radial)
		# Store whether the user wants the front or back of the sphere.
		sight = self.getRenderer3D().getCamera().getDirection()
		self.dragIsBackside = dotProduct(sight, radial) > 0.0
	
	# !! doesn't work excellently; imprecision leads to no xyzPrev
	def defaultDragHandler3D(self, xyStarting, xyPrevious, xyCurrent, name, xyzStarting):
		"""Following up on defaultBeginDragHandler3D(), lets the user rotate and translate 3D bodies around the scene."""
		# Get the transformation to the local coordinates of the picked body.
		body = self.getRenderer3D().getBodies()[name].getBody()
		center = body.getPosition()
		camera = self.getRenderer3D().getCamera()
		# Find the previous pick point on the sphere.
		(posPrev, dirPrev) = camera.getPickLine(xyPrevious)
		xyzPrev = lineSphereIntersection(posPrev, dirPrev, center, self.dragRadiusSquared, self.dragIsBackside)
		if xyzPrev != None:
			# Find the current pick point on the sphere.
			(posCurr, dirCurr) = camera.getPickLine(xyCurrent)
			xyzCurr = lineSphereIntersection(posCurr, dirCurr, center, self.dragRadiusSquared, self.dragIsBackside)
			if xyzCurr == None:
				# Construct the translation taking xyPrev to xyCurr.
				sight = camera.getDirection()
				xyzCurr = linePlaneIntersection(posCurr, dirCurr, xyzPrev, sight)
				if xyzCurr != None:
					delta = vectorDifference(xyzCurr, xyzPrev)
					body.setPosition(vectorSum(delta, center))
					self.needsRedrawing()
			else:
				# Construct rotation taking xyzPrev to xyzCurr.
				unitPrev = unitVector(vectorDifference(xyzPrev, center))
				unitCurr = unitVector(vectorDifference(xyzCurr, center))
				delta = quaternionFromVectors(unitPrev, unitCurr)
				# Prepend this rotation to the body's.
				body.setQuaternion(quaternionProduct(delta, body.getQuaternion()))
				self.needsRedrawing()

class Body2DJr(Body2D):
	"""A two-dimensional shape with a prescribed overall color and/or texture. Usually you make a shape using one of the specific constructor functions (makeDisk, etc.)."""
	
	def __init__(self, mesh):
		"""Initializes the shape with the given 2D mesh of vertices and texture coordinates."""
		Body2D.__init__(self, mesh)
		self.setTexture(None)
		self.setColor((1.0, 1.0, 1.0, 1.0))
	
	def completeColor(self, color):
		"""Private. Given an incomplete color (float L, sequence L, LA, RGB, or RGBA), fills it out to an RGBA tuple."""
		if color == None:
			# The default color is opaque white.
			return (1.0, 1.0, 1.0, 1.0)
		elif type(color) == float:
			# Interpret the scalar as a grayscale intensity.
			return (color, color, color, 1.0)
		elif type(color) == list or type(color) == tuple:
			if len(color) == 0:
				# The default color is opaque white.
				return (1.0, 1.0, 1.0, 1.0)
			elif len(color) == 1:
				# Interpret color as a grayscale intensity.
				return (color[0], color[0], color[0], 1.0)
			elif len(color) == 2:
				# Interpret color as a grayscale intensity and opacity.
				return (color[0], color[0], color[0], color[1])
			elif len(color) == 3:
				# Interpret color as RGB.
				return (color[0], color[1], color[2], 1.0)
			else:
				# Interpret color as RGB and opacity.
				return (color[0], color[1], color[2], color[3])
		else:
			# The default color is opaque white.
			return (1.0, 1.0, 1.0, 1.0)
	
	def setColor(self, color):
		"""Sets the shape's color to the given RGBA value. That is, color should be a 4-tuple or 4-list of floats between 0.0 and 1.0, indicating the red, green, blue, and opacity. The default is opaque white, (1.0, 1.0, 1.0, 1.0)."""
		if len(self.getUniforms()) == 0:
			self.getUniforms().append(self.completeColor(color))
		else:
			self.getUniforms()[0] = self.completeColor(color)
	
	def getColor(self):
		"""Returns the color as an RGBA tuple."""
		return self.getUniforms()[0]
	
	def addColor(self, delta):
		"""Adds the given RGBA color tuple to the shape's color."""
		change = self.completeColor(delta)
		old = self.getColor()
		self.setColor((old[0] + change[0], old[1] + change[1], old[2] + change[2], old[3] + change[3]))
	
	def setTexture(self, texture):
		"""Sets the shape's texture to the given Texture object (or None). The default is None."""
		if texture == None:
			texture = getWindow().getWhiteTexture()
		if len(self.getTextures()) == 0:
			self.getTextures().append(texture)
		else:
			self.getTextures()[0] = texture
	
	def getTexture(self):
		"""Returns the shape's Texture object (or None)."""
		texture = self.getTextures()[0]
		if texture == getWindow().getWhiteTexture():
			return None
		else:
			return texture

"""Constructor Functions"""

def makeDisk(radius, numSides=32, numAnnuli=1):
	"""Returns a Body2DJr object that is a disk (a filled circle) of the given radius. The remaining parameters control the fineness of the mesh. The position of the Body2DJr object indicates the position of the center of the disk."""
	(triangles, vertices, texCoords) = diskMeshData(radius, numSides, numAnnuli)
	return Body2DJr(Mesh(triangles, [vertices, texCoords]))

def makeSector(radius, start, end, numSides=8, numAnnuli=1):
	"""Returns a Body2DJr object that is a sector (a wedge from a disk, like a pie slice) of the given radius, running from the start angle to the end angle (in radians, counterclockwise from the positive x-direction). The remaining parameters control the fineness of the mesh. The position of the Body2DJr object indicates the position of the center of the corresponding disk."""
	(triangles, vertices, texCoords) = sectorMeshData(radius, start, end, numSides, numAnnuli)
	return Body2DJr(Mesh(triangles, [vertices, texCoords]))

def makeAnnulus(innerRadius, outerRadius, numSides=32, numAnnuli=1):
	"""Returns a Body2DJr object that is an annulus (a ring or washer) of the given inner and outer radii. The remaining parameters control the fineness of the mesh. The position of the Body2DJr object indicates the position of the center of the annulus."""
	(triangles, vertices, texCoords) = annulusMeshData(innerRadius, outerRadius, numSides, numAnnuli)
	return Body2DJr(Mesh(triangles, [vertices, texCoords]))

def makeRectangle(width, height, numWide=1, numHigh=1):
	"""Returns a Body2DJr object that is a rectangle of the given width and height. The remaining parameters control the fineness of the mesh. The position of the Body2DJr object indicates the position of the center of the rectangle."""
	(triangles, vertices, texCoords) = rectangleMeshData(width, height, numWide, numHigh)
	vertices = [(v[0] - 0.5 * width, v[1] - 0.5 * height) for v in vertices]
	return Body2DJr(Mesh(triangles, [vertices, texCoords]))

def makeRoundedRectangle(width, height, radius, numWide=1, numHigh=1, numSides=8, numAnnuli=1):
	"""Returns a Body2DJr object that is a rounded rectangle (like a playing card) of the given total width and height, with corners rounded off using circles of the given radius. The remaining parameters control the fineness of the mesh. The position of the Body2DJr object indicates the position of the center of the rounded rectangle."""
	(triangles, vertices, texCoords) = roundedRectangleMeshData(width, height, radius, numWide, numHigh, numSides, numAnnuli)
	vertices = [(v[0] - 0.5 * width, v[1] - 0.5 * height) for v in vertices]
	return Body2DJr(Mesh(triangles, [vertices, texCoords]))

def makeConvexPolygon(vertices):
	"""Returns a Body2DJr object that is a convex polygon, specified by its vertices in counterclockwise order. If the vertices describe a non-convex polygon, then unexpected visual artifacts may arise. To draw a non-convex polygon reliably, break it into convex pieces. The position of the Body2DJr object indicates the local origin, relative to which the vertices are specifed."""
	(triangles, vertices, texCoords) = convexPolygonMeshData(vertices)
	return Body2DJr(Mesh(triangles, [vertices, texCoords]))

def makeText(string, fontPath, fontSize):
	"""Returns a Body2DJr object that is a rectangle displaying a fixed string of text, drawn in a fixed TrueType font and size. On Mac OS X a typical fontPath is '/Library/Fonts/Silom.ttf'. A typical size is 12 or 18. The text occupies the shape's texture; setting the texture to something else obliterates the text. The text is opaque white on a transparent background; setting the shape's color therefore sets the text color. The position of the Body2DJr object indicates the position of the center of the text rectangle."""
	# Create the text texture.
	texture = makeTextTexture(string, fontPath, fontSize)
	# Construct the rectangle mesh data.
	wh = texture.getWidthHeight()
	width = float(wh[0])
	height = float(wh[1])
	(triangles, vertices, texCoords) = rectangleMeshData(width, height)
	# Adjust the texture coordinates, because the text is in the bottom-left corner of the texture.
	wh = texture.getBigWidthHeight()
	a = width / wh[0]
	b = height / wh[1]
	texCoords = [(0.0, 0.0), (0.0, b), (a, 0.0), (a, b)]
	vertices = [(v[0] - 0.5 * width, v[1] - 0.5 * height) for v in vertices]
	rectangle = Body2DJr(Mesh(triangles, [vertices, texCoords]))
	rectangle.setTexture(texture)
	return rectangle

class Body3DJr(Body3D):
	"""A three-dimensional solid with a prescribed overall color and/or texture. Usually you make a solid using one of the specific constructor functions (makeSphere, etc.)."""
	
	def __init__(self, mesh):
		"""Initializes the solid with the given 3D mesh of vertices, texture coordinates, and normals."""
		Body3D.__init__(self, mesh)
		self.setTexture(None)
		self.setColor((1.0, 1.0, 1.0, 1.0))
	
	def completeColor(self, color):
		"""Private method used to validate colors. Given an incomplete color (float L, sequence L, LA, RGB, or RGBA), fills it out to an RGBA tuple with A = 1.0."""
		if color == None:
			# The default color is opaque white.
			return (1.0, 1.0, 1.0, 1.0)
		elif type(color) == float:
			# Interpret the scalar as a grayscale intensity.
			return (color, color, color, 1.0)
		elif type(color) == list or type(color) == tuple:
			if len(color) == 0:
				# The default color is opaque white.
				return (1.0, 1.0, 1.0, 1.0)
			elif len(color) == 1:
				# Interpret color as a grayscale intensity.
				return (color[0], color[0], color[0], 1.0)
			elif len(color) == 2:
				# Interpret color as a grayscale intensity and (ignored) opacity.
				return (color[0], color[0], color[0], 1.0)
			elif len(color) == 3:
				# Interpret color as RGB.
				return (color[0], color[1], color[2], 1.0)
			else:
				# Interpret color as RGB and (ignored) opacity.
				return (color[0], color[1], color[2], 1.0)
		else:
			# The default color is opaque white.
			return (1.0, 1.0, 1.0, 1.0)
	
	def setColor(self, color):
		"""Sets the shape's color to the given RGB value. That is, color should be a 3-tuple or 3-list of floats between 0.0 and 1.0, indicating the red, green, and blue. Opacity is not supported. The default is white, (1.0, 1.0, 1.0)."""
		if len(self.getUniforms()) == 0:
			self.getUniforms().append(self.completeColor(color))
		else:
			self.getUniforms()[0] = self.completeColor(color)
	
	def getColor(self):
		"""Returns the color as an RGB tuple."""
		rgba = self.getUniforms()[0]
		return (rgba[0], rgba[1], rgba[2])
	
	def addColor(self, delta):
		"""Adds the given RGB color tuple to the shape's color."""
		change = self.completeColor(delta)
		old = self.getColor()
		self.setColor((old[0] + change[0], old[1] + change[1], old[2] + change[2]))
	
	def setTexture(self, texture):
		"""Sets the shape's texture to the given Texture object (or None). The default is None."""
		if texture == None:
			texture = getWindow().getWhiteTexture()
		if len(self.getTextures()) == 0:
			self.getTextures().append(texture)
		else:
			self.getTextures()[0] = texture
	
	def getTexture(self):
		"""Returns the shape's Texture object (or None)."""
		texture = self.getTextures()[0]
		if texture == getWindow().getWhiteTexture():
			return None
		else:
			return texture

"""Constructor Functions"""

def makeSphere(radius, numSides=32, numLayers=16):
	"""Returns a Body3DJr object that is a sphere of the given radius. The remaining parameters control the quality of the model. The position of the Body3DJr indicates the center of the sphere."""
	(triangles, vertices, texCoords, normals) = sphereMeshData(radius, numSides, numLayers)
	return Body3DJr(Mesh(triangles, [vertices, texCoords, normals]))

def makeEllipsoid(a, b, c, numSides=32, numLayers=16):
	"""Returns a Body3DJr object that is an ellipsoid of radii a, b, c in the x-, y-, z-directions, respectively. The remaining parameters control the quality of the model. The position of the Body3DJr indicates the center of the ellipsoid."""
	(triangles, vertices, texCoords, normals) = sphereMeshData(1.0, numSides, numLayers)
	(vertices, normals) = transformedMeshData(vertices, normals, (a, 0, 0, 0, b, 0, 0, 0, c))
	return Body3DJr(Mesh(triangles, [vertices, texCoords, normals]))

def makeBox(width, height, depth, numWide=1, numHigh=1, numDeep=1):
	"""Returns a Body3DJr object that is a width x height x depth box. The remaining parameters control the quality of the model. The position of the Body3DJr indicates the center of the box."""
	(triangles, vertices, texCoords, normals) = boxMeshData(width, height, depth, numWide, numHigh, numDeep)
	vertices = [(v[0] - 0.5 * width, v[1] - 0.5 * height, v[2] - 0.5 * depth) for v in vertices]
	return Body3DJr(Mesh(triangles, [vertices, texCoords, normals]))

def makeCylinder(height, radius, numSides=32, numLayers=1, numAnnuli=1):
	"""Returns a Body3DJr object that is a cylinder of the given height and radius. The ends of the cylinder are capped with disks. The remaining parameters control the quality of the model. The position of the Body3DJr indicates the center of the cylinder."""
	(triangles, vertices, texCoords, normals) = cylinderMeshData(height, radius, numSides, numLayers, numAnnuli)
	vertices = [(v[0], v[1], v[2] - 0.5 * height) for v in vertices]
	return Body3DJr(Mesh(triangles, [vertices, texCoords, normals]))

def makeCapsule(height, radius, numSides=32, numLayers=1, numAnnuli=8):
	"""Returns a Body3DJr object that is a capsule of the given height and radius. Specifically, height is the height of the cylindrical part of the capsule; the radius is the radius of the cylindrical part and of the hemispherical caps. The total height of the capsule is therefore height + 2.0 * radius. The remaining parameters control the quality of the model. The position of the Body3DJr indicates the center of the capsule."""
	(triangles, vertices, texCoords, normals) = capsuleMeshData(height, radius, numSides, numLayers, numAnnuli)
	vertices = [(v[0], v[1], v[2] - 0.5 * height) for v in vertices]
	return Body3DJr(Mesh(triangles, [vertices, texCoords, normals]))

def makeCone(height, radius, numSides=32, numLayers=1, numAnnuli=1):
	"""Returns a Body3DJr object that is a cone of the given height and radius. The end of the cone is capped with a disk. The remaining parameters control the quality of the model. The cone points up in the z-direction. The position of the Body3DJr indicates the center of mass of the cone; this is 2^(-1/3) ~ 0.7937 of the way down the cone."""
	(triangles, vertices, texCoords, normals) = coneMeshData(height, radius, numSides, numLayers, numAnnuli)
	shift = (1.0 - 2.0 ** (-1.0 / 3.0)) * height
	vertices = [(v[0], v[1], v[2] - shift) for v in vertices]
	return Body3DJr(Mesh(triangles, [vertices, texCoords, normals]))

def makeAxes(length, radius):
	"""Returns a triple of Body3DJrs. Each solid is a cylinder of the given length and radius. The cylinders point in the directions of the x-, y-, and z-axes, and are color-coded red, green, and blue."""
	xAxis = makeCylinder(length, radius)
	xAxis.setColor((1.0, 0.0, 0.0))
	xAxis.rotate(math.pi / 2.0, (0.0, 1.0, 0.0))
	xAxis.getBody().setPosition((0.5 * length, 0.0, 0.0))
	mesh = xAxis.getMesh()
	yAxis = Body3DJr(mesh)
	yAxis.setColor((0.0, 1.0, 0.0))
	yAxis.rotate(-math.pi / 2.0, (1.0, 0.0, 0.0))
	yAxis.getBody().setPosition((0.0, 0.5 * length, 0.0))
	zAxis = Body3DJr(mesh)
	zAxis.setColor((0.0, 0.0, 1.0))
	zAxis.getBody().setPosition((0.0, 0.0, 0.5 * length))
	return (xAxis, yAxis, zAxis)

def makeLandscape(width, height, elevations):
	"""Returns a pair (triangles, vertices, texCoords, normals) that can be used to build a 3D triangular mesh of a surface based on elevations. The surface's shadow in the x-y-plane is width x height. elevations is a 2D array of floats indicating the elevation of the surface above the x-y-plane at even spacings."""
	(triangles, vertices, texCoords, normals) = elevationMeshData(width, height, elevations)
	vertices = [(v[0] - 0.5 * width, v[1] - 0.5 * height, v[2]) for v in vertices]
	return Body3DJr(Mesh(triangles, [vertices, texCoords, normals]))

if __name__ == "__main__":
	# All Bopagopa programs begin by creating a window.
	window = WindowJr()
	# Make some 2D overlays.
	disk = makeDisk(64.0)
	disk.setColor((1.0, 0.0, 0.0, 1.0))
	disk.setPosition((100.0, 200.0))
	window.appendBody2D(disk)
	rectangle = makeRectangle(90.0, 70.0)
	rectangle.setColor((0.3, 0.8, 0.8, 0.7))
	rectangle.setPosition((300.0, 100.0))
	window.appendBody2D(rectangle)
	text = jr.makeText("Bopagopa Jr.", "/Library/Fonts/Silom.ttf", 18)
	text.setColor((1.0, 1.0, 1.0, 1.0))
	text.setPosition((128.0, 16.0))
	window.appendBody2D(text)
	# Make some 3D bodies.
	sphere = makeSphere(1.0)
	sphere.setColor((1.0, 1.0, 0.0))
	window.appendBody3D(sphere)
	cylinder = makeCylinder(1.0, 0.5)
	cylinder.setColor((0.0, 0.0, 1.0))
	cylinder.getBody().setPosition((1.0, 0.0, 1.0))
	window.appendBody3D(cylinder)
	capsule = makeCapsule(1.0, 0.5)
	capsule.setColor((1.0, 0.0, 1.0))
	capsule.getBody().setPosition((0.0, 1.0, 1.0))
	window.appendBody3D(capsule)
	box = makeBox(10.0, 10.0, 1.0)
	box.setColor((0.0, 1.0, 0.0))
	box.getBody().setPosition((0.0, 0.0, -1.0))
	window.appendBody3D(box)
	cone = makeCone(1.0, 0.5)
	cone.setColor((1.0, 0.5, 0.0))
	cone.getBody().setPosition((1.0, 1.0, 1.0))
	window.appendBody3D(cone)
	# Make some lights.
	light = SpotLight(512)
	window.setLights((light,))
	# All Bopagopa programs end by telling the window to run the user interface.
	beginUserInterface()
