# 2010 July 30

import OpenGL.GL as gl
from OpenGL.arrays import vbo
import numpy
from utility import *

class Mesh(object):
	"""An immutable triangular mesh built from vertex attribute arrays and triangles (triples of indices into the vertex attribute arrays). The attribute arrays might store 2D or 3D vertices, normals, texture coordinates, colors, or other material properties. However, Mesh does not have any sense of such semantics. It is simply a data repository. The semantics are declared in the renderer."""
	
	def __init__(self, triangles, attributes):
		"""Initializes the mesh. When this method is invoked, OpenGL must be ready to receive commands. attributes is a sequence of one or more attribute arrays, all of the same length. Each attribute array is a sequence of float-tuples, with all of the tuples in a given attribute array of the same length. For example, attributes might contain two elements, the first being a sequence of 4035 3-tuples (3D vertices, perhaps) and the second being a sequence of 4035 2-tuples (texture coordinates). triangles is a sequence of 3-tuples of unsigned integers serving as triangles into the attribute arrays."""
		# Store the size of the mesh.
		self.numTriangles = len(triangles)
		self.numVertices = len(attributes[0])
		self.attributeSizes = tuple([len(attribute[0]) for attribute in attributes])
		# Store the triangle triangles into an OpenGL vertex buffer object.
		if self.numVertices < 2**15:
			self.triangles = vbo.VBO(numpy.array(triangles, dtype=numpy.int16), 'GL_STATIC_DRAW', 'GL_ELEMENT_ARRAY_BUFFER')
		else:
			self.triangles = vbo.VBO(numpy.array(triangles, dtype=numpy.int32), 'GL_STATIC_DRAW', 'GL_ELEMENT_ARRAY_BUFFER')
		# Store the interleaved vertex attributes into an OpenGL vertex buffer object.
		interleave = []
		for attribute in attributes:
			interleave.append(numpy.array(attribute, dtype=numpy.float32))
		concatenated = numpy.concatenate(interleave, axis = 1)
		self.attributes = vbo.VBO(concatenated, 'GL_STATIC_DRAW', 'GL_ARRAY_BUFFER')
		# Store the stride and offsets for the interleaved vertex attribute VBO.
		self.stride = 4 * sum(self.attributeSizes)
		self.offsets = []
		for i in range(len(self.attributeSizes)):
			self.offsets.append(4 * sum(self.attributeSizes[:i]))
	
	def getNumTriangles(self):
		"""Returns the number of triangles."""
		return self.numTriangles
	
	def getNumVertices(self):
		"""Returns the number of vertices --- that is, the length of every attribute array."""
		return self.numVertices
	
	def getAttributeSizes(self):
		"""Returns a tuple of integers indicating how many floats are in each attribute. The length of this tuple is the number of attributes. The sum of this tuple is the total number of floats needed to describe the attributes of each vertex."""
		return self.attributeSizes
	
	# !! how much unbinding is necessary?
	def draw(self, locations):
		"""Non-user. Draws the mesh using the renderer's attribute locations. The attribute locations typically correspond one-to-one to the mesh's attributes, but in some cases the renderer may choose to pass an empty or partial list of locations."""
		self.attributes.bind()
		for i in range(len(locations)):
			if locations[i] != -1:
				gl.glVertexAttribPointer(locations[i], self.attributeSizes[i], gl.GL_FLOAT, False, self.stride, self.attributes + self.offsets[i])
		self.triangles.bind()
		if self.numVertices < 2**15:
			gl.glDrawElements(gl.GL_TRIANGLES, 3 * self.numTriangles, gl.GL_UNSIGNED_SHORT, self.triangles)
		else:
			gl.glDrawElements(gl.GL_TRIANGLES, 3 * self.numTriangles, gl.GL_UNSIGNED_INT, self.triangles)
		self.triangles.unbind()
		self.attributes.unbind()

"""Utility Functions"""

def newIndex(i, toCollapse, collapseTo):
	"""Non-user. Helper function for collapsedMeshData()."""
	try:
		# Recursively collapse this index, if appropriate.
		index = toCollapse.index(i)
		return newIndex(collapseTo[index], toCollapse, collapseTo)
	except ValueError:
		# This is not an index to collapse; decrement it instead.
		count = 0
		j = len(toCollapse) - 1
		while j >= 0 and toCollapse[j] < i:
			count += 1
			j -= 1
		return i - count

def collapsedMeshData(triangles, vertices, others=None):
	"""Returns new (triangles, vertices, others) that have been optimized by repeatedly collapsing equal vertices into each other (and adjusting the triangles accordingly). If others is not None, then it must be a list of arrays, each array of length equal to vertices, containing other attribute data (normals, colors, texture coordinates, etc.); whenever two vertices are collapsed in the new vertices, the corresponding entries in each array of others are collapsed in the new others."""
	# Figure out which vertices should collapse to which.
	toCollapse = []
	collapseTo = []
	for i in range(len(vertices) - 1, -1, -1):
		j = 0
		while j < i and vertices[i] != vertices[j]:
			j += 1
		if j < i:
			toCollapse.append(i)
			collapseTo.append(j)
	if len(toCollapse) == 0:
		# The number of vertices has stabilized; return.
		if others == None:
			return (triangles, vertices)
		else:
			return (triangles, vertices, others)
	else:
		# Construct new vertices and others.
		newVertices = []
		if others == None:
			newOthers = None
		else:
			newOthers = [[] for other in others]
		for i in range(len(vertices)):
			if i not in toCollapse:
				# Copy vertex and other attributes into new lists.
				newVertices.append(vertices[i])
				if others != None:
					for j in range(len(others)):
						newOthers[j].append(others[j][i])
		# Construct new triangles.
		newIndices = []
		for tri in triangles:
			newTri = (newIndex(tri[0], toCollapse, collapseTo), newIndex(tri[1], toCollapse, collapseTo), newIndex(tri[2], toCollapse, collapseTo))
			if newTri[0] != newTri[1] and newTri[1] != newTri[2] and newTri[2] != newTri[0]:
				newIndices.append(newTri)
		# Recur in case there are more vertices to collapse.
		return collapsedMeshData(newIndices, newVertices, newOthers)

def closeButDistinct(v, vertices, epsilon):
	"""Non-user helper function for equatedMeshVertices(). If v is within epsilon of any of the vertices, as measured by the l-infinity norm, but distinct from that vertex, then returns the index of that vertex; otherwise returns -1."""
	for i in range(len(vertices)):
		equal = True
		distant = False
		for k in range(len(v)):
			diff = abs(vertices[i][k] - v[k])
			if diff >= epsilon:
				distant = True
			elif diff > 0:
				equal = False
		if not distant and not equal:
			return i
	return -1

def equatedVertices(vertices, epsilon):
	"""Returns a new list of vertices, of the same length as the old list, such that all vertices within epsilon of each other, as measured by the l-infinity norm, have been adjusted to be equal. Used to clean up after precision errors. Consider using this function and then collapedVertices."""
	changed = False
	newVertices = []
	for v in vertices:
		index = closeButDistinct(v, newVertices, epsilon)
		if index >= 0:
			newVertices.append(newVertices[index])
			changed = True
		else:
			newVertices.append(v)
	if changed:
		return equatedVertices(newVertices, epsilon)
	else:
		return vertices

def joinedMeshData(meshDataList):
	"""Given a sequence meshDataList of sequences (triangles, attrib1, attrib2, ...), returns a single sequence [triangles, attrib1, attrib2, ...] that joins them all together, so that a joined mesh can be made."""
	joined = []
	for i in range(len(meshDataList[0])):
		joined.append([])
	offset = 0
	for meshData in meshDataList:
		joined[0].extend([(tri[0] + offset, tri[1] + offset, tri[2] + offset) for tri in meshData[0]])
		offset += len(meshData[1])
		for i in range(1, len(meshData)):
			joined[i].extend(meshData[i])
	return joined

def nontrivialTriangles(triangles, vertices):
	"""Returns a new list of triangles, that is identical to the old list with all trivial triangles removed. A triangle is trivial if two of its vertices are identical."""
	new = []
	for i in range(len(triangles)):
		if vertices[triangles[i][0]] != vertices[triangles[i][1]] and vertices[triangles[i][1]] != vertices[triangles[i][2]] and vertices[triangles[i][2]] != vertices[triangles[i][0]]:
			new.append(triangles[i])
	return new

def smoothNormals(triangles, vertices):
	"""Returns a list of unit normals, one per vertex, formed by averaging the true unit normals of all adjacent triangles."""
	normals = [(0.0, 0.0, 0.0)] * len(vertices)
	for tri in triangles:
		# Compute the triangle's true unit normal vector.
		v = vectorDifference(vertices[tri[2]], vertices[tri[1]])
		w = vectorDifference(vertices[tri[0]], vertices[tri[1]])
		n = unitVector(crossProduct(v, w))
		# Add this normal to the normal at each of the triangle's vertices.
		for j in range(3):
			normals[tri[j]] = vectorSum(n, normals[tri[j]])
	# Normalize the normal vector at each vertex.
	for i in range(len(normals)):
		normals[i] = unitVector(normals[i])
	return normals

def transformedMeshData(vertices, normals, m):
	"""Given 3D triangular mesh data and a 3x3 matrix m (given as a 9-sequence in row-major order), returns a pair (vertices, normals) that describes the original mesh transformed by the matrix. That is, vertices transform by m, and normals transform by the inverse transpose of m (and are then renormalized). The matrix must be invertible."""
	newVertices = [matrixProduct(m, v, 3, 3, 1) for v in vertices]
	mTI = matrixInverse(matrixTranspose(m, 3, 3))
	newNormals = [nestNormalized(matrixProduct(mTI, n, 3, 3, 1)) for n in normals]
	return (newVertices, newNormals)

def deformedMeshData(vertices, normals, function, jacobian):
	"""The inputs vertices and normals are 3D vertices and unit normals at those vertices. The input function is a function that takes in a 3-tuple of floats and outputs a 3-tuple of floats; that is, it deforms 3D space. The input jacobian is a function that takes in a 3-tuple of floats and outputs a 9-tuple of floats; it should be the Jacobian matrix of function (in row-major order). Returns a pair (vertices, normals) that describes the original mesh deformed by the function. That is, vertices transform by function, and normals transform by the inverse transpose of the Jacobian (and are then normalized). In the special case where the Jacobian is constant and function is the corresponding linear transformation, deformedMeshData is equivalent to transformedMeshData. The Jacobian must be invertible at all of the vertices."""
	newVertices = []
	newNormals = []
	for i in range(len(vertices)):
		newVertices.append(function(vertices[i]))
		jTI = matrixInverse3(matrixTranspose(jacobian(vertices[i]), 3, 3))
		newNormals.append(nestNormalized(matrixProduct(jTI, normals[i], 3, 3, 1)))
	return (newVertices, newNormals)

def localDeformer(displacement, center, radius, cutoff, cutoffDerivative):
	"""The inputs displacement and center are 3-sequences of floats, radius is a positive float, cutoff is a cutoff function on the interval [0.0, 1.0], and cutoffDerivative is its derivative. (The Bopagopa utility functions cubicCutoff and cubicCutoffDerivative are examples.) Returns a pair (function, jacobian) suitable for use as the deformer in deformedMeshData. The deformer leaves points more than radius away from the center unmoved. It moves other points based on their distance from the center --- specifically, by cutoff(distance^2 / radius^2) * displacement. Assuming that cutoff(0.0) is 1.0 and cutoff(1.0) is 0.0, the deformer's effect is to move the center by displacement, and to adjust a neighborhood around the center accordingly."""
	# Define the main deformer function.
	def function(x):
		d = nestDifference(x, center)
		dSquared = nestDot(d, d) / (radius * radius)
		if dSquared >= 1.0:
			return x
		else:
			return nestSum(x, nestScaling(cutoff(dSquared), displacement))
	# Define its Jacobian matrix function.
	def jacobian(x):
		d = nestDifference(x, center)
		dSquared = nestDot(d, d) / (radius * radius)
		if dSquared >= 1.0:
			return (1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0)
		else:
			coefficient = 2.0 * cutoffDerivative(dSquared) / (radius * radius)
			j00 = displacement[0] * coefficient * d[0] + 1.0
			j01 = displacement[0] * coefficient * d[1]
			j02 = displacement[0] * coefficient * d[2]
			j10 = displacement[1] * coefficient * d[0]
			j11 = displacement[1] * coefficient * d[1] + 1.0
			j12 = displacement[1] * coefficient * d[2]
			j20 = displacement[2] * coefficient * d[0]
			j21 = displacement[2] * coefficient * d[1]
			j22 = displacement[2] * coefficient * d[2] + 1.0
			return (j00, j01, j02, j10, j11, j12, j20, j21, j22)
	return (function, jacobian)

"""2D Constructor Functions"""

def diskMeshData(radius, numSides=16, numAnnuli=1):
	"""Returns (indices, vertices, texCoords) for a 2D triangular mesh of a disk of the given radius. Uses numSides subdivisions around and numAnnuli subdivisions out. The center of the disk is at the origin."""
	# Construct the vertices.
	vertices = [(0.0, 0.0)]
	for i in range(numSides):
		a = i * 2.0 * math.pi / numSides
		x = math.cos(a)
		y = math.sin(a)
		for j in range(numAnnuli):
			r = radius * (j + 1) / numAnnuli
			vertices.append((r * x, r * y))
	# Construct the indices.
	indices = []
	for i in range(numSides - 1):
		indices.append((0, i * numAnnuli + 1, (i + 1) * numAnnuli + 1))
		for j in range(1, numAnnuli):
			indices.append((i * numAnnuli + j, i * numAnnuli + j + 1, (i + 1) * numAnnuli + j + 1))
			indices.append((i * numAnnuli + j, (i + 1) * numAnnuli + j + 1, (i + 1) * numAnnuli + j))
	indices.append((0, (numSides - 1) * numAnnuli + 1, 1))
	for j in range(1, numAnnuli):
		indices.append(((numSides - 1) * numAnnuli + j, (numSides - 1) * numAnnuli + j + 1, j + 1))
		indices.append(((numSides - 1) * numAnnuli + j, j + 1, j))
	# Construct the texture coordinates.
	texCoords = [(v[0] * 0.5 / radius + 0.5, v[1] * 0.5 / radius + 0.5) for v in vertices]
	return (indices, vertices, texCoords)

def sectorMeshData(radius, start, end, numSides=8, numAnnuli=1):
	"""Returns (indices, vertices) for a 2D triangular mesh of a sector of a disk of the given radius, running between the angles start and end (in radians, counterclockwise from the positive x-axis). Uses numSides subdivisions around and numAnnuli subdivisions out. The center of the sector (that is, the center of the corresponding disk) is at the origin."""
	# Construct the vertices.
	vertices = [(0.0, 0.0)]
	for i in range(numSides + 1):
		a = start + i * (end - start) / numSides
		x = math.cos(a)
		y = math.sin(a)
		for j in range(numAnnuli):
			r = radius * (j + 1) / numAnnuli
			vertices.append((r * x, r * y))
	# Construct the indices.
	indices = []
	for i in range(numSides):
		indices.append((0, i * numAnnuli + 1, (i + 1) * numAnnuli + 1))
		for j in range(1, numAnnuli):
			indices.append((i * numAnnuli + j, i * numAnnuli + j + 1, (i + 1) * numAnnuli + j + 1))
			indices.append((i * numAnnuli + j, (i + 1) * numAnnuli + j + 1, (i + 1) * numAnnuli + j))
	# Construct the texture coordinates.
	texCoords = [(v[0] * 0.5 / radius + 0.5, v[1] * 0.5 / radius + 0.5) for v in vertices]
	return (indices, vertices, texCoords)

def annulusMeshData(inner, outer, numSides=16, numAnnuli=1):
	"""Returns (indices, vertices, texCoords) for a 2D triangular mesh of an annulus of the given radii. Uses numSides subdivisions around and numAnnuli subdivisions out. The center of the annulus is at the origin."""
	# Construct vertices.
	vertices = []
	for i in range(numSides):
		a = i * 2.0 * math.pi / numSides
		x = math.cos(a)
		y = math.sin(a)
		for j in range(numAnnuli + 1):
			r = inner + (outer - inner) * j / numAnnuli
			vertices.append((r * x, r * y))
	# Construct indices.
	indices = []
	for i in range(numSides - 1):
		for j in range(numAnnuli):
			indices.append((i * (numAnnuli + 1) + j, i * (numAnnuli + 1) + j + 1, (i + 1) * (numAnnuli + 1) + j + 1))
			indices.append((i * (numAnnuli + 1) + j, (i + 1) * (numAnnuli + 1) + j + 1, (i + 1) * (numAnnuli + 1) + j))
	for j in range(numAnnuli):
		indices.append(((numSides - 1) * (numAnnuli + 1) + j, (numSides - 1) * (numAnnuli + 1) + j + 1, j + 1))
		indices.append(((numSides - 1) * (numAnnuli + 1) + j, j + 1, j))
	# Construct the texture coordinates.
	texCoords = [(v[0] * 0.5 / outer + 0.5, v[1] * 0.5 / outer + 0.5) for v in vertices]
	return (indices, vertices, texCoords)

def rectangleMeshData(width, height, numWide=1, numHigh=1):
	"""Returns (indices, vertices, texCoords) for a 2D triangular mesh of a rectangle, of the given width and length. Uses numWide by numHigh subdivisions. The bottom left corner of the rectangle is at the origin."""
	# Build the indices.
	indices = []
	for i in range(numWide):
		for j in range(numHigh):
			indices.append((i * (numHigh + 1) + j, (i + 1) * (numHigh + 1) + j, (i + 1) * (numHigh + 1) + j + 1))
			indices.append((i * (numHigh + 1) + j, (i + 1) * (numHigh + 1) + j + 1, i * (numHigh + 1) + j + 1))
	# Build the vertices.
	xDelta = width / numWide
	yDelta = height / numHigh
	vertices = []
	# Compute most of the vertices.
	x = 0.0
	for i in range(numWide):
		y = 0.0
		for j in range(numHigh):
			vertices.append((x, y))
			y += yDelta
		# Exactly compute the last vertex in this column.
		vertices.append((x, height))
		x += xDelta
	# Exactly compute the last column of vertices.
	y = 0.0
	for j in range(numHigh):
		vertices.append((width, y))
		y += yDelta
	# Exactly compute the last vertex.
	vertices.append((width, height))
	# Construct the texture coordinates.
	size = max(width, height)
	texCoords = [(v[0] / size, v[1] / size) for v in vertices]
	return (indices, vertices, texCoords)

def roundedRectangleMeshData(width, height, radius, numWide=1, numHigh=1, numSides=4, numAnnuli=1):
	"""Returns (indices, vertices, texCoords) for a 2D triangular mesh of a rectangle with rounded corners, of given overall width and height and given radius of rounded corners. For the core rectangle uses numWide by numHigh subdivisions; for each quarter disk uses numSides subdivisions around and numAnnuli subdivisions out. The bottom left corner (of the bounding non-rounded rectangle) is at the origin."""
	# Build nine pieces separately.
	(trInds, trVerts, temp) = sectorMeshData(radius, 0.0, 0.5 * math.pi, numSides, numAnnuli)
	trVerts = [(v[0] + width - radius, v[1] + height - radius) for v in trVerts]
	(tlInds, tlVerts, temp) = sectorMeshData(radius, 0.5 * math.pi, math.pi, numSides, numAnnuli)
	tlVerts = [(v[0] + radius, v[1] + height - radius) for v in tlVerts]
	(blInds, blVerts, temp) = sectorMeshData(radius, math.pi, 1.5 * math.pi, numSides, numAnnuli)
	blVerts = [(v[0] + radius, v[1] + radius) for v in blVerts]
	(brInds, brVerts, temp) = sectorMeshData(radius, 1.5 * math.pi, 2.0 * math.pi, numSides, numAnnuli)
	brVerts = [(v[0] + width - radius, v[1] + radius) for v in brVerts]
	(lInds, lVerts, temp) = rectangleMeshData(radius, height - 2.0 * radius, numAnnuli, numHigh)
	lVerts = [(v[0], v[1] + radius) for v in lVerts]
	rInds = [i for i in lInds]
	rVerts = [(v[0] + width - radius, v[1]) for v in lVerts]
	(bInds, bVerts, temp) = rectangleMeshData(width - 2.0 * radius, radius, numWide, numAnnuli)
	bVerts = [(v[0] + radius, v[1]) for v in bVerts]
	tInds = [i for i in bInds]
	tVerts = [(v[0], v[1] + height - radius) for v in bVerts]
	(cInds, cVerts, temp) = rectangleMeshData(width - 2.0 * radius, height - 2.0 * radius, numWide, numHigh)
	cVerts = [(v[0] + radius, v[1] + radius) for v in cVerts]
	# Join them into a single mesh.
	(indices, vertices) = joinedMeshData([(trInds, trVerts), (tlInds, tlVerts), (blInds, blVerts), (brInds, brVerts), (rInds, rVerts), (tInds, tVerts), (lInds, lVerts), (bInds, bVerts), (cInds, cVerts)])
	# Equate vertices that are close together.
	vertices = equatedVertices(vertices, radius / (4.0 * numAnnuli))
	# Collapse vertices that are equal.
	(indices, vertices) = collapsedMeshData(indices, vertices)
	# Construct the texture coordinates.
	size = max(width, height)
	texCoords = [(v[0] / size, v[1] / size) for v in vertices]
	return (indices, vertices, texCoords)

def convexPolygonMeshData(vertices):
	"""Returns (indices, vertices, texCoords) for a 2D triangular mesh of a polygon with the given vertices in counterclockwise order. If the polygon is not convex then strange visual artifacts may arise. The positions of the vertices are all relative to the origin."""
	# The vertices are given; construct the indices.
	indices = []
	for i in range(1, len(vertices) - 1):
		indices.append((0, i, i + 1))
	# Construct the texture coordinates.
	minX = min([v[0] for v in vertices])
	maxX = max([v[0] for v in vertices])
	minY = min([v[1] for v in vertices])
	maxY = max([v[1] for v in vertices])
	if maxX - minX > maxY - minY:
		texCoords = [((v[0] - minX) / (maxX - minX), (v[1] - minY) / (maxX - minX)) for v in vertices]
	else:
		texCoords = [((v[0] - minX) / (maxY - minY), (v[1] - minY) / (maxY - minY)) for v in vertices]
	return (indices, vertices, texCoords)

"""3D Constructor Functions"""

def elevationMeshData(width, height, elevations):
	"""Returns a pair (triangles, vertices, texCoords, normals) that can be used to build a 3D triangular mesh of a surface based on elevations. The surface's shadow is width x height. elevations is a 2D array of floats indicating the elevation of the surface above the z = 0 plane at even spacings."""
	(triangles, vertices, texCoords) = rectangleMeshData(width, height, len(elevations) - 1, len(elevations[0]) - 1)
	for i in range(len(elevations)):
		for j in range(len(elevations[0])):
			k = i * len(elevations[0]) + j
			vertices[k] = (vertices[k][0], vertices[k][1], elevations[i][j])
	normals = smoothNormals(triangles, vertices)
	return (triangles, vertices, texCoords, normals)

def boxMeshData(width, height, depth, numWide=1, numHigh=1, numDeep=1):
	"""Returns a tuple (triangles, vertices, texCoords, normals) that can be used to build a 3D triangular mesh of a box. The edges of the box come with redundant vertices because the normals change abruptly, and to facilitate texture mapping. The box extends from the origin into the first octant of space."""
	# Build the sides in the z = 0 and z = depth planes.
	(inds, verts, texs) = rectangleMeshData(width, height, numWide, numHigh)
	topVertices = [(v[0], v[1], depth) for v in verts]
	topIndices = inds
	topTexCoords = texs
	normals = [(0.0, 0.0, 1.0) for v in verts]
	bottomVertices = [(v[0], height - v[1], 0.0) for v in verts]
	bottomIndices = [i for i in inds]
	bottomTexCoords = [t for t in texs]
	normals.extend([(0.0, 0.0, -1.0) for v in verts])
	# Build the sides in the y = 0 and y = height planes.
	(inds, verts, texs) = rectangleMeshData(width, depth, numWide, numDeep)
	frontVertices = [(v[0], 0.0, v[1]) for v in verts]
	frontIndices = inds
	frontTexCoords = texs
	normals.extend([(0.0, -1.0, 0.0) for v in verts])
	backVertices = [(width - v[0], height, v[1]) for v in verts]
	backIndices = [i for i in inds]
	backTexCoords = [t for t in texs]
	normals.extend([(0.0, 1.0, 0.0) for v in verts])
	# Build the sides in the x = 0 and x = width planes.
	(inds, verts, texs) = rectangleMeshData(depth, height, numDeep, numHigh)
	leftVertices = [(0.0, v[1], v[0]) for v in verts]
	leftIndices = inds
	leftTexCoords = texs
	normals.extend([(-1.0, 0.0, 0.0) for v in verts])
	rightVertices = [(width, height - v[1], v[0]) for v in verts]
	rightIndices = [i for i in inds]
	rightTexCoords = [t for t in texs]
	normals.extend([(1.0, 0.0, 0.0) for v in verts])
	# Join together all six parts; the normals are already joined.
	(triangles, vertices, texCoords) = joinedMeshData(((topIndices, topVertices, topTexCoords), (bottomIndices, bottomVertices, bottomTexCoords), (frontIndices, frontVertices, frontTexCoords), (backIndices, backVertices, backTexCoords), (leftIndices, leftVertices, leftTexCoords), (rightIndices, rightVertices, rightTexCoords)))
	epsilon = min(width, height, depth) / (4.0 * max(numWide, numHigh, numDeep))
	vertices = equatedVertices(vertices, epsilon)
	return (triangles, vertices, texCoords, normals)

def surfaceOfRevolutionMeshData(zs, rs, numSides=16, numAnnuli=1):
	"""Returns a tuple (triangle, vertices, texCoords, normals) that can be used to build a mesh for a surface of revolution. The surface is formed by revolving a piecewise-linear curve about the z-axis. The arguments zs and rs are sequences of floats of equal length at least 2. The vertices of the curve have z-values given by zs, and nonnegative radii (out from the z-axis) given by rs. The first z should be less than the last z, to guarantee correct orientation. If the first radius is not 0.0, then that end of the solid is capped with a disk; the same goes for the last radius. You can depend on the vertices occuring in a particular order: by meridian, from the zs[0] end of the meridian to the zs[-1] end of the meridian."""
	# Cap the bottom with a disk.
	if rs[0] == 0.0:
		bottomIndices = []
		bottomVertices = []
		bottomTexCoords = []
		bottomNormals = []
	else:
		(bottomIndices, bottomVertices, bottomTexCoords) = diskMeshData(rs[0], numSides, numAnnuli)
		bottomVertices = [(v[0], -v[1], zs[0]) for v in bottomVertices]
		bottomNormals = [(0.0, 0.0, -1.0)] * len(bottomVertices)
	# Construct the triangles and texture coordinates of the side of the surface.
	(sideIndices, vertices, sideTexCoords) = rectangleMeshData(1.0, 1.0, numSides, len(zs) - 1)
	# Construct the first vertex on the prime meridian.
	vertices = [(rs[0], 0.0, zs[0])]
	normals = [unitVector((zs[1] - zs[0], 0.0, rs[0] - rs[1]))]
	# Construct most of the prime meridian.
	for i in range(1, len(zs) - 1):
		vertices.append((rs[i], 0.0, zs[i]))
		normals.append(unitVector((zs[i + 1] - zs[i - 1], 0.0, rs[i - 1] - rs[i + 1])))
	# Construct the last vertex on the prime meridian.
	vertices.append((rs[-1], 0.0, zs[-1]))
	normals.append(unitVector((zs[-1] - zs[-2], 0.0, rs[-2] - rs[-1])))
	# Start the side vertices with those on the prime meridian.
	sideVertices = []
	sideNormals = []
	sideVertices.extend(vertices)
	sideNormals.extend(normals)
	# Continue the side vertices by rotating the prime meridian about the z-axis.
	delta = 2.0 * math.pi / numSides
	for i in range(1, numSides):
		cosine = math.cos(i * delta)
		sine = math.sin(i * delta)
		for j in range(len(vertices)):
			sideVertices.append((cosine * vertices[j][0], sine * vertices[j][0], vertices[j][2]))
			sideNormals.append((cosine * normals[j][0], sine * normals[j][0], normals[j][2]))
	# Finish the side vertices with those on the prime meridian again.
	sideVertices.extend(vertices)
	sideNormals.extend(normals)
	# Cap the top with a disk.
	if rs[-1] == 0.0:
		topIndices = []
		topVertices = []
		topTexCoords = []
		topNormals = []
	else:
		(topIndices, topVertices, topTexCoords) = diskMeshData(rs[-1], numSides, numAnnuli)
		topVertices = [(v[0], v[1], zs[-1]) for v in topVertices]
		topNormals = [(0.0, 0.0, 1.0)] * len(topVertices)
	# Join together the three parts of the surface.
	(triangles, vertices, texCoords, normals) = joinedMeshData(((bottomIndices, bottomVertices, bottomTexCoords, bottomNormals), (sideIndices, sideVertices, sideTexCoords, sideNormals), (topIndices, topVertices, topTexCoords, topNormals)))
	# Compute the minimum legitimate distance between two distinct vertices, divided by 4.0.
	minParallel = min(rs) * 2.0 * math.pi / numSides
	minMeridian = (zs[1] - zs[0])**2 + (rs[1] - rs[0])**2
	for i in range(1, len(zs) - 1):
		minMeridian = min(minMeridian, (zs[i + 1] - zs[i])**2 + (rs[i + 1] - rs[i])**2)
	minMeridian = math.sqrt(minMeridian)
	epsilon = min(minParallel, minMeridian) / 4.0
	# Clean up precision errors and trivial triangles.
	vertices = equatedVertices(vertices, epsilon)
	triangles = nontrivialTriangles(triangles, vertices)
	return (triangles, vertices, texCoords, normals)

def sphereMeshData(radius, numSides=16, numLayers=8):
	"""Returns a tuple (triangles, vertices, texCoords, normals) that can be used to build a sphere mesh of the given radius (centered at the origin)."""
	# Construct the south pole.
	zs = [-radius]
	rs = [0.0]
	# Construct most of the prime meridian.
	delta = math.pi / numLayers
	for i in range(1, numLayers):
		zs.append(-radius * math.cos(i * delta))
		rs.append(radius * math.sin(i * delta))
	# Construct the north pole.
	zs.append(radius)
	rs.append(0.0)
	# Construct the surface of revolution.
	(triangles, vertices, texCoords, normals) = surfaceOfRevolutionMeshData(zs, rs, numSides)
	# Clean up the normals at the south and north poles.
	for i in range(numSides + 1):
		normals[i * (numLayers + 1)] = (0.0, 0.0, -1.0)
		normals[(i + 1) * (numLayers + 1) - 1] = (0.0, 0.0, 1.0)
	return (triangles, vertices, texCoords, normals)

def cylinderMeshData(height, radius, numSides=16, numLayers=1, numAnnuli=1):
	"""Returns a tuple (triangles, vertices, texCoords, normals) that can be used to build a 3D triangular mesh of a cylinder. The cylinder extends from the origin along the positive z-axis."""
	zs = [i * height / numLayers for i in range(numLayers)]
	zs.append(height)
	rs = [radius] * (numLayers + 1)
	return surfaceOfRevolutionMeshData(zs, rs, numSides, numAnnuli)

def capsuleMeshData(height, radius, numSides=16, numLayers=8, numAnnuli=4):
	"""Returns a tuple (triangles, vertices, texCoords, normals) that can be used to build a 3D triangular mesh of a capsule (a cylinder capped with hemispheres). Specifically, height is the height of the cylindrical part of the capsule; the radius is the radius of the cylindrical part and of the hemispherical caps. The total height of the capsule is therefore height + 2.0 * radius. The cylindrical part of the capsule extends from the origin along the positive z-axis."""
	# Construct the south pole.
	zs = [-radius]
	rs = [0.0]
	# Construct most of the prime meridian of the lower hemisphere.
	delta = 0.5 * math.pi / numAnnuli
	for i in range(1, numAnnuli):
		zs.append(-radius * math.cos(i * delta))
		rs.append(radius * math.sin(i * delta))
	# Construct the prime meridian of the cylinder.
	for i in range(numLayers + 1):
		zs.append(i * height / numLayers)
		rs.append(radius)
	# Construct most of the prime meridian of the upper hemisphere.
	for i in range(numAnnuli + 1, 2 * numAnnuli):
		zs.append(-radius * math.cos(i * delta) + height)
		rs.append(radius * math.sin(i * delta))
	# Construct the north pole.
	zs.append(radius + height)
	rs.append(0.0)
	# Construct the surface of revolution.
	(triangles, vertices, texCoords, normals) = surfaceOfRevolutionMeshData(zs, rs, numSides)
	# Clean up the normals at the south and north poles.
	for i in range(numSides + 1):
		normals[i * (numLayers + 2 * numAnnuli + 1)] = (0.0, 0.0, -1.0)
		normals[(i + 1) * (numLayers + 2 * numAnnuli + 1) - 1] = (0.0, 0.0, 1.0)
	return (triangles, vertices, texCoords, normals)

def coneMeshData(height, radius, numSides=16, numLayers=1, numAnnuli=1):
	"""Returns a tuple (triangles, vertices, texCoords, normals) that can be used to build a 3D triangular mesh of a cone. The cone extends from the origin along the positive z-axis, and points up."""
	zs = [i * height / numLayers for i in range(numLayers)]
	zs.append(height)
	rs = [-i * radius / numLayers + radius for i in range(numLayers)]
	rs.append(0.0)
	return surfaceOfRevolutionMeshData(zs, rs, numSides, numAnnuli)

