# 2010 August 18 by Joshua R. Davis

import Image
import ImageFont
import OpenGL.GL as gl

class Texture(object):
	"""A 2D texture to be applied to 2D and 3D bodies. Usually one builds a texture using the constructor functions below, rather than using the class' constructor method directly."""
	
	def __init__(self, data, width, height, isRepeating):
		"""The data argument is a string of bytes, of length = width x height x numChannels, in row-major order. The width and height need not be powers of 2, but the image will be padded out as if they were powers of 2. The isRepeating Boolean indicates how texture coordinates beyond [0, 1] (and at the boundary) are handled; either by repeating the texture or clamping to the edge value."""
		# Store some basic information.
		self.width = width
		self.height = height
		self.numChannels = len(data) / (width * height)
		if self.numChannels == 1:
			self.format = gl.GL_LUMINANCE
		elif self.numChannels == 2:
			self.format = gl.GL_LUMINANCE_ALPHA
		elif self.numChannels == 3:
			self.format = gl.GL_RGB
		elif self.numChannels == 4:
			self.format = gl.GL_RGBA
		else:
			raise TypeError("Texture data have " + self.numChannels + " channels.")
		# Pad the width and height out to powers of 2.
		self.bigWidth = 1
		while self.bigWidth < self.width:
			self.bigWidth *= 2
		self.bigHeight = 1
		while self.bigHeight < self.height:
			self.bigHeight *= 2
		# Recopy the data, if necessary.
		if self.width != self.bigWidth or self.height != self.bigHeight:
			original = data
			data = ""
			for j in range(self.height):
				raster = ""
				for i in range(self.width):
					offset = self.numChannels * (j * self.width + i)
					pixel = ""
					for k in range(self.numChannels):
						pixel += original[offset + k]
					raster += pixel
				for i in range(self.width, self.bigWidth):
					raster += pixel
				data += raster
			for j in range(self.height, self.bigHeight):
				data += raster
		# Create an OpenGL texture object.
		self.texture = gl.glGenTextures(1)
		gl.glBindTexture(gl.GL_TEXTURE_2D, self.texture)
		gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)
		gl.glPixelStorei(gl.GL_PACK_ALIGNMENT, 1)
		if isRepeating:
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_REPEAT)
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_REPEAT)
		else:
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_EDGE)
			gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP_TO_EDGE)
		gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
		gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
		# Load the data into the texture.
		gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, self.numChannels, self.bigWidth, self.bigHeight, 0, self.format, gl.GL_UNSIGNED_BYTE, data)
	
	def __del__(self):
		"""Non-user. Cleans up the underlying OpenGL texture object."""
		gl.glDeleteTextures(self.texture)
		# The parent class object does not have __del__().
		#object.__del__(self)
	
	def getBigWidthHeight(self):
		"""Returns the internal width and height in pixels. These are integers, always powers of 2 --- namely, the first powers of 2 greater than or equal to the width and height."""
		return (self.bigWidth, self.bigHeight)
	
	def getWidthHeight(self):
		"""Returns the width and height in pixels. These are integers, not necessarily powers of 2."""
		return (self.width, self.height)
	
	def getNumChannels(self):
		"""Returns the number of channels per pixel."""
		return self.numChannels
	
	def getFormat(self):
		"""Returns GL_LUMINANCE, GL_LUMINANCE_ALPHA, GL_RGB, or GL_RGBA (which are defined in the module OpenGL.GL)."""
		return self.format
	
	def bind(self):
		"""Non-user. Binds the texture to the active texture unit, so that it affects subsequent OpenGL drawing."""
		gl.glBindTexture(gl.GL_TEXTURE_2D, self.texture)
		gl.glEnable(gl.GL_TEXTURE_2D)

"""Constructor Functions"""

def makeFileTexture(path, isRepeating=False):
	"""Returns a Texture object initialized from a graphics file at the given path. The isRepeating Boolean indicates how texture coordinates beyond [0, 1] (and at the boundary) are handled --- either by repeating the texture or clamping to the edge value. If the file cannot be found, then None is returned."""
	try:
		image = Image.open(path)
	except IOError:
		print "The file " + path + " could not be loaded."
		return None
	# If the image is paletted, then convert it to RGB or whatever it's supposed to be.
	if image.mode == 'P':
		image = image.convert(image.palette.mode)
	data = image.tostring("raw", image.mode, 0, -1)
	return Texture(data, image.size[0], image.size[1], isRepeating)

def makeTextTexture(string, fontPath, fontSize):
	"""Returns a Texture object containing an image of a string of text, drawn in a given TrueType font and size. On Mac OS X a typical fontPath is '/Library/Fonts/Silom.ttf'. A typical size is 12 or 18. In the text image, the text is opaque white, the background is transparent white, and antialiasing is accomplished by opacity."""
	try:
		font = ImageFont.truetype(fontPath, fontSize)
	except IOError:
		print "The font " + fontPath + " could not be found in size " + str(fontSize) + "."
		return None
	bitmap = font.getmask(string)
	# Pad the width and height out to powers of 2.
	bigWidth = 1
	while bigWidth < bitmap.size[0]:
		bigWidth *= 2
	bigHeight = 1
	while bigHeight < bitmap.size[1]:
		bigHeight *= 2
	# Laboriously copy the bitmap, flipped, into a buffer string.
	data = ""
	for j in range(bitmap.size[1] - 1, -1, -1):
		raster = ""
		for i in range(bitmap.size[0]):
			pixel = chr(255) + chr(bitmap.getpixel((i, j)))
			raster += pixel
		data += raster
	return Texture(data, bitmap.size[0], bitmap.size[1], False)

def makeWhiteTexture():
	"""Returns a 1x1 opaque white texture."""
	return Texture(2 * chr(255), 1, 1, True)

