# 2010 October 3 by Joshua R. Davis

import math
import OpenGL.GL as gl
import OpenGL.GLUT as glut

def beginUserInterface():
	"""Starts the user interface running. Call this after the initial scene is set up. After this, all user events are handled by the window(s)."""
	glut.glutMainLoop()

def compiledShader(sourceCode, shaderType):
	"""Non-user. Helper function for compiledShaderProgram(). Given a non-empty string of GLSL shader code and a type (GL_VERTEX_SHADER or GL_FRAGMENT_SHADER), returns the compiled shader."""
	# This code is adapted from PyOpenGL 3.0.1.
	if isinstance(sourceCode, (str, unicode)):
		sourceCode = [sourceCode]
	shader = gl.glCreateShader(shaderType)
	gl.glShaderSource(shader, sourceCode)
	gl.glCompileShader(shader)
	result = gl.glGetShaderiv(shader, gl.GL_COMPILE_STATUS)
	if not(result):
		# TODO: this will be wrong if the user has disabled traditional unpacking array support.
		raise RuntimeError(
			"Shader compile failure (%s): %s"%(
				result,
				gl.glGetShaderInfoLog(shader),
			),
			sourceCode,
			shaderType,
		)
	return shader
	
def compiledShaderProgram(vertexCode, fragmentCode):
	"""Non-user. Given two non-empty strings representing an OpenGL Shading Language vertex shader and fragment shader, compiles and links them into a shader program. Returns the shader program identifier. You can subsequently set uniforms using glUniform*(), check the validity/completeness of the program using validateShaderProgram(), and use the program using glUseProgram(). When you are finished with the program, call glDeleteProgram()."""
	program = gl.glCreateProgram()
	vertexShader = compiledShader(vertexCode, gl.GL_VERTEX_SHADER)
	gl.glAttachShader(program, vertexShader)
	fragmentShader = compiledShader(fragmentCode, gl.GL_FRAGMENT_SHADER)
	gl.glAttachShader(program, fragmentShader)
	gl.glLinkProgram(program)
	# Mark the shaders for deletion when the program gets deleted.
	gl.glDeleteShader(vertexShader)
	gl.glDeleteShader(fragmentShader)
	return program

def validateShaderProgram(shaderProgram):
	"""Non-user. Validates the compiled and linked shader program, raising RuntimeError if something goes wrong."""
	gl.glValidateProgram(shaderProgram)
	validation = gl.glGetProgramiv(shaderProgram, gl.GL_VALIDATE_STATUS)
	if validation == gl.GL_FALSE:
		raise RuntimeError(
			"Validation failure (%s): %s"%(
			validation,
			gl.glGetProgramInfoLog(shaderProgram),
		))
	link_status = gl.glGetProgramiv(shaderProgram, gl.GL_LINK_STATUS)
	if link_status == gl.GL_FALSE:
		raise RuntimeError(
			"Link failure (%s): %s"%(
			link_status,
			gl.glGetProgramInfoLog(shaderProgram),
		))

def nestSum(a, b):
	"""Returns the sum a + b of two nested lists of numbers, assuming that they have the same structure."""
	if type(a) == list or type(a) == tuple:
		result = []
		for i in range(len(a)):
			result.append(nestSum(a[i], b[i]))
		return result
	else:
		return a + b

def nestDifference(a, b):
	"""Returns the difference a - b of two nested lists of numbers, assuming that they have the same structure."""
	if type(a) == list or type(a) == tuple:
		result = []
		for i in range(len(a)):
			result.append(nestDifference(a[i], b[i]))
		return result
	else:
		return a - b

def nestScaling(t, a):
	"""Returns the scaling t * a of a nested list of numbers a by a scalar t."""
	if type(a) == list or type(a) == tuple:
		result = []
		for i in range(len(a)):
			result.append(nestScaling(t, a[i]))
		return result
	else:
		return t * a

def nestCombination(a, t, b):
	"""Returns the combination a + t * b of two nested lists of numbers, assuming that they have the same structure."""
	if type(a) == list or type(a) == tuple:
		result = []
		for i in range(len(a)):
			result.append(nestCombination(a[i], t, b[i]))
		return result
	else:
		return a + t * b

def nestInterpolation(a, t, b):
	"""Returns the interpolation a + t * (b - a) of two nested lists of numbers, assuming that they have the same structure."""
	if type(a) == list or type(a) == tuple:
		result = []
		for i in range(len(a)):
			result.append(nestInterpolation(a[i], t, b[i]))
		return result
	else:
		return a + t * (b - a)

def nestDot(a, b):
	"""Returns the dot product a.b of two nested lists of numbers, assuming that they have the same structure."""
	if type(a) == list or type(a) == tuple:
		result = 0.0
		for i in range(len(a)):
			result += nestDot(a[i], b[i])
		return result
	else:
		return a * b

def nestNorm(a):
	"""Returns the square root of the dot product of the nested list of numbers with itself. For example, if the nest is a vector, then returns the length of the vector."""
	return math.sqrt(nestDot(a, a))

def nestNormalized(a):
	"""Returns the normalization of the nest a --- that is, the nest of the same structure, whose overall Euclidean norm is 1, and whose elements have the same signs as the elements of a do."""
	normSquared = nestDot(a, a)
	if normSquared == 0.0:
		return a
	else:
		return nestScaling(math.sqrt(1.0 / normSquared), a)

def vectorSum(v, w):
	"""Deprecated; use nest version. Returns the sum of 3-dimensional vectors."""
	return [v[0] + w[0], v[1] + w[1], v[2] + w[2]]

def vectorDifference(v, w):
	"""Deprecated; use nest version. Returns the difference v - w of 3-dimensional vectors."""
	return [v[0] - w[0], v[1] - w[1], v[2] - w[2]]

def vectorScaling(v, s):
	"""Deprecated; use nest version. Returns the 3-dimensional vector scaled by the given scalar."""
	return [s * v[0], s * v[1], s * v[2]]

def vectorNorm(v):
	"""Deprecated; use nest version. Returns the length of the 3-dimensional vector."""
	return math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])

def unitVector(v):
	"""Deprecated; use nest version. Returns the 3-dimensional vector scaled to unit length, or the zero vector."""
	norm = vectorNorm(v)
	if norm == 0.0:
		return v
	else:
		return vectorScaling(v, 1.0 / norm)

def dotProduct(v, w):
	"""Deprecated; use nest version. Returns the dot product of two 3-dimensional vectors."""
	return v[0] * w[0] + v[1] * w[1] + v[2] * w[2]

def crossProduct(v, w):
	"""Returns the cross product of two 3-dimensional vectors."""
	return [v[1] * w[2] - v[2] * w[1], v[2] * w[0] - v[0] * w[2], v[0] * w[1] - v[1] * w[0]]

def arccos(x):
	"""Clamps x to the interval [-1, 1], then takes the inverse cosine."""
	return math.acos(max(min(x, 1.0), -1.0))

def cartesianFromSpherical(rpt):
	"""Returns Cartesian coordinates [x, y, z] for spherical coordinates rpt = (rho, phi, theta). Typically 0 <= rho, 0 <= phi <= pi, 0 <= theta <= 2 * pi."""
	return [rpt[0] * math.sin(rpt[1]) * math.cos(rpt[2]), rpt[0] * math.sin(rpt[1]) * math.sin(rpt[2]), rpt[0] * math.cos(rpt[1])]

def sphericalFromCartesian(v):
	"""Returns spherical coordinates [rho, phi, theta] for Cartesian coordinates v = (x, y, z). Satisfies 0 <= rho, 0 <= phi <= pi, 0 <= theta <= 2 * pi."""
	# rho is the radius; normalize the vector.
	rho = nestNorm(v)
	if rho == 0.0:
		return [0.0, 0.0, 0.0]
	else:
		a = v[0] / rho
		b = v[1] / rho
		c = v[2] / rho
		phi = arccos(c)
		sinPhi = math.sin(phi)
		if sinPhi == 0.0:
			# The vector is vertical.
			if c > 0.0:
				return [rho, 0.0, 0.0]
			else:
				return [rho, math.pi / 2.0, 0.0]
		else:
			# This is the typical case.
			if b < 0.0:
				return [rho, phi, 2.0 * math.pi - arccos(a / sinPhi)]
			else:
				return [rho, phi, arccos(a / sinPhi)]

def quaternionProduct(a, b):
	"""Returns the product of two quaternions in rectangular [w, x, y, z] form."""
	w = a[0] * b[0] - a[1] * b[1] - a[2] * b[2] - a[3] * b[3]
	x = a[2] * b[3] - a[3] * b[2] + a[0] * b[1] + b[0] * a[1]
	y = a[3] * b[1] - a[1] * b[3] + a[0] * b[2] + b[0] * a[2]
	z = a[1] * b[2] - a[2] * b[1] + a[0] * b[3] + b[0] * a[3]
	return [w, x, y, z]

def quaternionFromAngleAxis(au):
	"""Given an angle-axis rotation [a, [u1, u2, u3]], returns the rotation as a quaternion [w, x, y, z]."""
	cosine = math.cos(0.5 * au[0])
	sine = math.sin(0.5 * au[0])
	return [cosine, sine * au[1][0], sine * au[1][1], sine * au[1][2]]

def angleAxisFromQuaternion(wxyz):
	"""Given a quaternion in the form [w, x, y, z], returns an angle-axis description [a, [u1, u2, u3]] of the same rotation."""
	norm = math.sqrt(wxyz[1] * wxyz[1] + wxyz[2] * wxyz[2] + wxyz[3] * wxyz[3])
	try:
		recip = 1.0 / norm
		angle = 2.0 * arccos(wxyz[0])
		return [angle, [wxyz[1] * recip, wxyz[2] * recip, wxyz[3] * recip]]
	except ZeroDivisionError:
		return [0.0, [0.0, 0.0, 1.0]]

def angleFromAxisAndVectors(u, v, w):
	"""Returns the angle in radians, such that a rotation about u takes v to w. All input vectors must be unit."""
	if nestDot(u, crossProduct(v, w)) < 0.0:
		return -arccos(nestDot(v, w))
	else:
		return arccos(nestDot(v, w))

def matrixInverse3(a):
	"""Returns the inverse of the given 3x3 matrix a, or None if it is not invertible. All matrices are given as flat sequences in row-major order."""
	# Compute the first column of the adjoint of a.
	b00 = a[1 * 3 + 1] * a[2 * 3 + 2] - a[1 * 3 + 2] * a[2 * 3 + 1]
	b10 = a[1 * 3 + 2] * a[2 * 3 + 0] - a[1 * 3 + 0] * a[2 * 3 + 2]
	b20 = a[1 * 3 + 0] * a[2 * 3 + 1] - a[1 * 3 + 1] * a[2 * 3 + 0]
	# Compute the determinant of a by expanding along the first column.
	determinant = a[0 * 3 + 0] * b00 + a[0 * 3 + 1] * b10 + a[0 * 3 + 2] * b20
	if determinant == 0.0:
		return None
	else:
		reciprocal = 1.0 / determinant
		# Compute the first column of the inverse.
		b00 *= reciprocal
		b10 *= reciprocal
		b20 *= reciprocal
		# Compute the second column of the inverse.
		b01 = (a[0 * 3 + 2] * a[2 * 3 + 1] - a[0 * 3 + 1] * a[2 * 3 + 2]) * reciprocal
		b11 = (a[0 * 3 + 0] * a[2 * 3 + 2] - a[0 * 3 + 2] * a[2 * 3 + 0]) * reciprocal
		b21 = (a[0 * 3 + 1] * a[2 * 3 + 0] - a[0 * 3 + 0] * a[2 * 3 + 1]) * reciprocal
		# Compute the third column of the inverse.
		b02 = (a[0 * 3 + 1] * a[1 * 3 + 2] - a[0 * 3 + 2] * a[1 * 3 + 1]) * reciprocal
		b12 = (a[0 * 3 + 2] * a[1 * 3 + 0] - a[0 * 3 + 0] * a[1 * 3 + 2]) * reciprocal
		b22 = (a[0 * 3 + 0] * a[1 * 3 + 1] - a[0 * 3 + 1] * a[1 * 3 + 0]) * reciprocal
		return (b00, b01, b02, b10, b22, b12, b20, b21, b22)

def matrixTranspose(a, m, n):
	"""Returns the nxm transpose of the given mxn matrix a. All matrices are given as flat sequences in row-major order. The entries in the matrix can be of any type, and of mixed types."""
	result = []
	for i in range(n):
		for j in range(m):
			result.append(a[j * n + i])
	return result

def matrixProduct(a, b, m, n, p):
	"""Returns the product of an mxn matrix a and an nxp matrix b, as an mxp matrix. All matrices are given as flat sequences in row-major order."""
	result = []
	for i in range(m):
		for j in range(p):
			entry = 0
			for k in range(n):
				entry += a[i * n + k] * b[k * p + j]
			result.append(entry)
	return result

def affineMatrixProduct(a, b):
	"""Takes as input two 4x4 affine matrices --- that is, matrices with last row (0.0, 0.0, 0.0, 1.0) --- as length-16 sequences in row-major order. Returns their matrix product, in the same format."""
	ab00 = a[0 * 4 + 0] * b[0 * 4 + 0] + a[0 * 4 + 1] * b[1 * 4 + 0] + a[0 * 4 + 2] * b[2 * 4 + 0]
	ab01 = a[0 * 4 + 0] * b[0 * 4 + 1] + a[0 * 4 + 1] * b[1 * 4 + 1] + a[0 * 4 + 2] * b[2 * 4 + 1]
	ab02 = a[0 * 4 + 0] * b[0 * 4 + 2] + a[0 * 4 + 1] * b[1 * 4 + 2] + a[0 * 4 + 2] * b[2 * 4 + 2]
	ab03 = a[0 * 4 + 0] * b[0 * 4 + 3] + a[0 * 4 + 1] * b[1 * 4 + 3] + a[0 * 4 + 2] * b[2 * 4 + 3] + a[0 * 4 + 3]
	ab10 = a[1 * 4 + 0] * b[0 * 4 + 0] + a[1 * 4 + 1] * b[1 * 4 + 0] + a[1 * 4 + 2] * b[2 * 4 + 0]
	ab11 = a[1 * 4 + 0] * b[0 * 4 + 1] + a[1 * 4 + 1] * b[1 * 4 + 1] + a[1 * 4 + 2] * b[2 * 4 + 1]
	ab12 = a[1 * 4 + 0] * b[0 * 4 + 2] + a[1 * 4 + 1] * b[1 * 4 + 2] + a[1 * 4 + 2] * b[2 * 4 + 2]
	ab13 = a[1 * 4 + 0] * b[0 * 4 + 3] + a[1 * 4 + 1] * b[1 * 4 + 3] + a[1 * 4 + 2] * b[2 * 4 + 3] + a[1 * 4 + 3]
	ab20 = a[2 * 4 + 0] * b[0 * 4 + 0] + a[2 * 4 + 1] * b[1 * 4 + 0] + a[2 * 4 + 2] * b[2 * 4 + 0]
	ab21 = a[2 * 4 + 0] * b[0 * 4 + 1] + a[2 * 4 + 1] * b[1 * 4 + 1] + a[2 * 4 + 2] * b[2 * 4 + 1]
	ab22 = a[2 * 4 + 0] * b[0 * 4 + 2] + a[2 * 4 + 1] * b[1 * 4 + 2] + a[2 * 4 + 2] * b[2 * 4 + 2]
	ab23 = a[2 * 4 + 0] * b[0 * 4 + 3] + a[2 * 4 + 1] * b[1 * 4 + 3] + a[2 * 4 + 2] * b[2 * 4 + 3] + a[2 * 4 + 3]
	return (ab00, ab01, ab02, ab03, ab10, ab11, ab12, ab13, ab20, ab21, ab22, ab23, 0.0, 0.0, 0.0, 1.0)

def perpendicularUnitVector(v):
	"""Given a unit vector, returns a unit vector perpendicular to it."""
	if v[0] >= v[1]:
		if v[1] >= v[2]:
			# v[2] is smallest.
			w = (-v[1], v[0], 0.0)
		else:
			# v[1] is smallest.
			w = (-v[2], 0.0, v[0])
	else:
		if v[0] >= v[2]:
			# v[2] is smallest.
			w = (-v[1], v[0], 0.0)
		else:
			# v[0] is smallest.
			w = (0.0, -v[2], v[1])
	return nestNormalized(w)

def angleAxisFromVectors(v, w):
	"""Given two unit vectors v and w, returns the angle-axis rotation about v x w that takes v to w. If v and w are parallel, then v x w is ill-defined, and this function chooses it arbitrarily."""
	u = crossProduct(v, w)
	dot = nestDot(v, w)
	if nestDot(u, u) == 0.0:
		if dot < 0.0:
			# v and w are opposite.
			u = unitPerpendicularVector(v)
			return [math.pi, [u[0], u[1], u[2]]]
		else:
			# v and w are identical.
			return [0.0, [0.0, 0.0, 1.0]]
	else:
		# This is the typical case.
		u = nestNormalized(u)
		return [arccos(dot), [u[0], u[1], u[2]]]

def quaternionFromVectors(v, w):
	"""Given two unit vectors v and w, returns the quaternion representing the rotation about v x w that takes v to w. If v and w are parallel, then v x w is ill-defined, and this function chooses it to be [0, 0, 1] arbitrarily."""
	vdotw = nestDot(v, w)
	u = crossProduct(v, w)
	udotu = nestDot(u, u)
	if udotu == 0.0:
		if vdotw < 0.0:
			# v and w are opposite.
			return [0.0, 0.0, 0.0, 1.0]
		else:
			# v and w are identical.
			return [1.0, 0.0, 0.0, 0.0]
	else:
		u = nestScaling(1.0 / math.sqrt(udotu), u)
		# This is the typical case.
		cosine = math.sqrt(0.5 * (1.0 + vdotw))
		sine = math.sqrt(0.5 * (1.0 - vdotw))
		return [cosine, sine * u[0], sine * u[1], sine * u[2]]

def vectorRotated(au, v):
	"""Returns the vector v rotated by the angle-axis rotation au."""
	wxyz = quaternionFromAngleAxis(au)
	xyz = wxyz[1:]
	cross = crossProduct(xyz, v)
	term1 = nestScaling(-2.0, crossProduct(cross, xyz))
	term2 = nestScaling(2.0 * wxyz[0], cross)
	return nestSum(nestSum(term1, term2), v)

def quadraticSolutions(a, b, c):
	"""Returns the real solutions to the quadratic polynomial a t^2 + b t + c = 0. If a = b = c = 0 then all t are solutions; this is signaled by returning None. Otherwise a list of 0, 1, or 2 numbers is returned. If 2 are returned, then the first is less than the second."""
	if a == 0.0:
		if b == 0.0:
			if c == 0.0:
				return None
			else:
				return []
		else:
			return [-c / b]
	else:
		discriminant = b * b - 4.0 * a * c
		if discriminant < 0.0:
			return []
		else:
			root = math.sqrt(discriminant)
			if root == 0.0:
				return [-b / (2.0 * a)]
			else:
				return [(-b - root) / (2.0 * a), (-b + root) / (2.0 * a)]

def lineSphereIntersection(a, v, c, rr, backside=False):
	"""Intersects a line and a sphere. The line is described parametrically by a + t * v. The sphere is described by its center c and radius-squared rr. The backside parameter indicates whether the first (False) or second (True) solution is desired. If there is just one solution, then backside has no effect. If there is no solution, then None is returned."""
	ca = nestDifference(a, c)
	t = quadraticSolutions(nestDot(v, v), 2.0 * nestDot(ca, v), nestDot(ca, ca) - rr)
	if t == None or len(t) == 0:
		return None
	else:
		if backside:
			t = t[-1]
		else:
			t = t[0]
		return nestSum(a, nestScaling(t, v))

def linePlaneIntersection(a, v, c, n):
	"""Intersects a line and a plane. The line is described parametrically by a + t * v. The plane is described by a point c and a normal vector n. If there is no solution, then None is returned."""
	ndotv = dotProduct(n, v)
	if ndotv == 0.0:
		return None
	else:
		t = nestDot(n, nestDifference(c, a)) / ndotv
		return nestSum(a, nestScaling(t, v))

def cubicBijection(t):
	"""A smooth, increasing, bijective function from the interval [0.0, 1.0] to itself, with derivative zero at both ends. Very similar to cubicCutoff. Useful for smooth animations."""
	return -2.0 * t * t * t + 3.0 * t * t

def cubicCutoff(d):
	"""A smooth, decreasing, bijective function from the interval [0.0, 1.0] to itself, with derivative zero at both ends. Very similar to cubicBijection. Useful for cutting off nonzero functions."""
	return 2.0 * d * d * d - 3 * d * d + 1.0

def cubicCutoffDerivative(d):
	"""The derivative of cubicCutoff with respect to its argument."""
	return 6.0 * d * (d - 1.0)


