# 2010 October 9 by Joshua R. Davis

"""
A basic version of Conway's Game of Life. It's not really a "game", but rather a cellular automaton --- but it's fun anyway. The game takes place on an infinite grid of square cells. At each step of the game, each cell is either alive or dead, depending on the previous step. Specifically, if a live cell has 2 or 3 live neighbords, then it stays alive; if it has fewer than 2 or more than 3 live neighbors, it dies. If a dead cell has exactly 3 live neighbors, then it becomes alive; otherwise, it remains dead.

Controls:
Space       Pause and unpause the game.
Return      Pause the game, and perform a single step of the game.
Left/Right  Slow down and speed up the game.
Up/Down     Zoom in and out.
Mouse       Click or drag to pan across the board (i.e. move your view).
Shift-Mouse Click or drag to turn cells on or off.
Mouse Menu  Insert a canonical example into the current game.
"""

import time
import math
import OpenGL.GLUT as glut
import bopagopa.bopagopajr as jr

class GameOfLife(object):
	"""Runs the game of life simulation. Responds to user interface events."""
	
	def __init__(self):
		"""Initializes a new game of life with an empty board."""
		# Build the infrastructure for creating cells.
		meshData = jr.rectangleMeshData(1.0, 1.0)
		self.mesh = jr.Mesh(meshData[0], (meshData[1], meshData[2]))
		self.spares = []
		# Register callbacks with the window.
		self.window = jr.getWindow()
		self.window.setKeyboardHandler(self.handleKeyboard)
		self.window.setBeginDragHandler2D(self.handleBeginDrag2D)
		self.window.setDragHandler2D(self.handleDrag2D)
		self.window.setBeginDragHandlerEmpty(self.handleBeginDrag)
		self.window.setDragHandlerEmpty(self.handleDrag)
		self.window.setClickHandler2D(self.handleClick2D)
		self.window.setClickHandlerEmpty(self.handleClick)
		self.window.setMenuHandler(self.handleMenu, ["Glider NE", "Acorn", "Gosper glider gun SE"])
		self.window.trackTime(self.passTime)
		# Set up instance variables.
		self.group = jr.Group2D()
		self.window.appendBody2D(self.group)
		self.stepTime = 0.5
		self.zoom = 8.0
		self.setZoom(8.0)
		self.board = {}
		self.pause()
		self.dragMode = "panning"
	
	def activateCell(self, ij):
		"""Assumes that the cell at ij was dead; makes it alive."""
		if len(self.spares) == 0:
			cell = jr.Body2DJr(self.mesh)
		else:
			cell = self.spares.pop()
		cell.setPosition((float(ij[0]), float(ij[1])))
		self.group.getBodies().append(cell)
		self.board[ij] = cell
	
	def setCell(self, ij, isAlive):
		"""Sets cell ij to be alive or not in the current dictionary, and updates the display."""
		wasAlive = ij in self.board
		if isAlive and not wasAlive:
			self.activateCell(ij)
			self.window.needsRedrawing()
		elif wasAlive and not isAlive:
			# Kill off a cell that was alive.
			cell = self.board[ij]
			self.spares.append(cell)
			del self.board[ij]
			self.group.getBodies().remove(cell)
			self.window.needsRedrawing()
	
	def incrementDictionary(self, dictionary, ij):
		"""In the given dictionary, increments the value associated to the key ij. If ij was not in the dictionary, then sets its value to be 1."""
		if ij in dictionary:
			dictionary[ij] += 1
		else:
			dictionary[ij] = 1
	
	def performStep(self):
		"""Performs a single step of the game of life."""
		# Detach all of the cell shapes from the window.
		bodies = self.group.getBodies()
		while len(bodies) > 0:
			self.spares.append(bodies.pop())
		# Compute a dictionary counting the number of living neighbors.
		dictionary = {}
		for ij in self.board:
			self.incrementDictionary(dictionary, (ij[0] + 1, ij[1]))
			self.incrementDictionary(dictionary, (ij[0] + 1, ij[1] + 1))
			self.incrementDictionary(dictionary, (ij[0], ij[1] + 1))
			self.incrementDictionary(dictionary, (ij[0] - 1, ij[1] + 1))
			self.incrementDictionary(dictionary, (ij[0] - 1, ij[1]))
			self.incrementDictionary(dictionary, (ij[0] - 1, ij[1] - 1))
			self.incrementDictionary(dictionary, (ij[0], ij[1] - 1))
			self.incrementDictionary(dictionary, (ij[0] + 1, ij[1] - 1))
		# Convert the count dictionary into the new board dictionary.
		oldBoard = self.board
		self.board = {}
		for ij in dictionary:
			wasAlive = ij in oldBoard
			count = dictionary[ij]
			if (wasAlive and (count == 2 or count == 3)) or (not wasAlive and count == 3):
				self.activateCell(ij)
		# Usually the game needs redrawing.
		self.window.needsRedrawing()
	
	def handleKeyboard(self, xy, key):
		"""Responds to keyboard events."""
		if key == ' ':
			if self.paused:
				self.unpause()
			else:
				self.pause()
		elif key == '\r':
			self.pause()
			self.performStep()
		elif key == glut.GLUT_KEY_LEFT:
			self.stepTime = min(1.0, self.stepTime + 0.1)
		elif key == glut.GLUT_KEY_RIGHT:
			self.stepTime = max(0.05, self.stepTime - 0.1)
		elif key == glut.GLUT_KEY_UP:
			self.setZoom(self.zoom + 1.0)
		elif key == glut.GLUT_KEY_DOWN:
			self.setZoom(self.zoom - 1.0)
		elif key == 'f':
			self.window.setIsFullScreen(not self.window.getIsFullScreen())
		elif key == 'w':
			self.window.setIsWireframe(not self.window.getIsWireframe())
	
	def setZoom(self, zoom):
		"""Sets the zoom level, clamping it between 1 and 8."""
		# Figure out which cell the display was centered on.
		center = (0.5 * self.window.getWidth(), 0.5 * self.window.getHeight())
		ij = self.getLocation(center)
		# Change the zoom level.
		self.zoom = min(8.0, max(1.0, zoom))
		self.group.setScales((self.zoom, self.zoom))
		# Recenter the display.
		xy = self.getPosition(ij)
		self.group.translate((center[0] - xy[0], center[1] - xy[1]))
		self.window.needsRedrawing()
	
	def handleMenu(self, string):
		"""Handles menu events."""
		xy = self.window.getMouse()
		ij = self.getLocation(xy)
		self.insertExample(string, ij)
	
	def handleBeginDrag(self, xyStarting, xyCurrent):
		"""Pans the board or starts laying down live cells."""
		if self.window.shiftKeyWasDown():
			self.dragMode = "activating"
			self.handleDrag(xyStarting, xyStarting, xyStarting)
		else:
			self.dragMode = "panning"
		self.handleDrag(xyStarting, xyStarting, xyCurrent)
	
	def handleBeginDrag2D(self, xyStarting, xyCurrent, name):
		"""Pans the board or starts laying down dead cells."""
		if self.window.shiftKeyWasDown():
			self.dragMode = "deactivating"
			self.handleDrag(xyStarting, xyStarting, xyStarting)
		else:
			self.dragMode = "panning"
		self.handleDrag(xyStarting, xyStarting, xyCurrent)
	
	def handleDrag(self, xyStarting, xyPrevious, xyCurrent):
		"""Continues panning the board or laying down cells."""
		if self.dragMode == "panning":
			delta = (xyCurrent[0] - xyPrevious[0], xyCurrent[1] - xyPrevious[1])
			self.group.translate(delta)
			self.window.needsRedrawing()
		else:
			ij = self.getLocation(xyCurrent)
			self.setCell(ij, self.dragMode == "activating")
	
	def handleDrag2D(self, xyStarting, xyPrevious, xyCurrent, name):
		"""Continues panning the board or laying down cells."""
		self.handleDrag(xyStarting, xyPrevious, xyCurrent)
	
	def handleClick(self, xy):
		"""Pans the board or lays down a live cell."""
		ij = self.getLocation(xy)
		if self.window.shiftKeyWasDown():
			self.setCell(ij, True)
		else:
			center = (0.5 * self.window.getWidth(), 0.5 * self.window.getHeight())
			delta = (center[0] - xy[0], center[1] - xy[1])
			self.window.makeDisplacer(0.1, self.group.translate, ((0.0, 0.0),), (delta,), 1.0, jr.cubicBijection)
	
	def handleClick2D(self, xy, name):
		"""Pans the board or kills off a live cell."""
		ij = self.getLocation(xy)
		if self.window.shiftKeyWasDown():
			self.setCell(ij, False)
		else:
			center = (0.5 * self.window.getWidth(), 0.5 * self.window.getHeight())
			delta = (center[0] - xy[0], center[1] - xy[1])
			self.window.makeDisplacer(0.1, self.group.translate, ((0.0, 0.0),), (delta,), 1.0, jr.cubicBijection)
	
	def passTime(self, delta):
		"""Tracks the passage of time, and performs a single game step at regular intervals."""
		if not self.paused:
			self.timePassed += delta
			if self.timePassed >= self.stepTime:
				self.timePassed = 0.0
				self.performStep()
	
	def pause(self):
		"""Pauses the game."""
		self.paused = True
	
	def unpause(self):
		"""Unpauses the game."""
		self.paused = False
		self.timePassed = 0.0
	
	def getLocation(self, xy):
		"""Returns the location (i, j) at the screen coordinates xy."""
		ab = self.group.getPosition()
		i = math.floor((xy[0] - ab[0]) / self.zoom)
		j = math.floor((xy[1] - ab[1]) / self.zoom)
		return (i, j)
	
	def getPosition(self, ij):
		"""Returns the bottom-left screen position (x, y) for the grid location ij."""
		ab = self.group.getPosition()
		return (ab[0] + self.zoom * ij[0], ab[1] + self.zoom * ij[1])
	
	def insertExample(self, name, ij):
		"""Near the location ij, turns on a bunch of cells according to a preset pattern."""
		if name == "Glider NE":
			pixels = [(0, 0), (0, -1), (0, -2), (-1, 0), (-2, -1)]
		elif name == "Acorn":
			pixels = [(0, 0), (1, 0), (1, 2), (3, 1), (4, 0), (5, 0), (6, 0)]
		elif name == "Gosper glider gun SE":
			pixels = [(0, 0), (-1, 0), (-1, 1), (-1, -1), (-2, 2), (-2, -2), (-3, 0), (-4, 3), (-4, -3), (-5, 3), (-5, -3), (-6, 2), (-6, -2), (-7, 1), (-7, -1), (-7, 0), (-16, 0), (-16, 1), (-17, 0), (-17, 1), (3, 1), (3, 2), (3, 3), (4, 1), (4, 2), (4, 3), (5, 0), (5, 4), (7, 0), (7, -1), (7, 4), (7, 5), (17, 2), (17, 3), (18, 2), (18, 3)]
		else:
			pixels = []
		for ab in pixels:
			self.setCell((ij[0] + ab[0], ij[1] + ab[1]), True)

if __name__ == "__main__":
	window = jr.WindowJr(512, 512, "Game of Life")
	game = GameOfLife()
	jr.beginUserInterface()
