# 2010 July 7 by Joshua R. Davis

"""
This program demonstrates how to use the three kinds of animations: Positioners, Displacers, and Delayers. It is similar to the preceding tutorial, but uses animations to make all changes smooth. The only "new feature" is that shift-clicking on a shape causes it to translate after a delay of two seconds. Animations are easy to use. Whenever they fire, the window automatically redraws itself. Whenever they are finished, the window automatically gets rid of them.
"""

import bopagopa.bopagopajr as jr
import math
import OpenGL.GLUT as glut

def handleKeyboard(xy, key):
	"""Translates and rotates shape 0, doing everything with animations that take half a second."""
	window = jr.getWindow()
	shape = window.getBodies2D()[0]
	if key == glut.GLUT_KEY_LEFT:
		# Instead of doing shape.translate((-10.0, 0.0)), we do this.
		window.makeDisplacer(0.5, shape.translate, ((0.0, 0.0),), ((-10.0, 0.0),))
	elif key == glut.GLUT_KEY_RIGHT:
		window.makeDisplacer(0.5, shape.translate, ((0.0, 0.0),), ((10.0, 0.0),))
	elif key == glut.GLUT_KEY_DOWN:
		window.makeDisplacer(0.5, shape.translate, ((0.0, 0.0),), ((0.0, -10.0),))
	elif key == glut.GLUT_KEY_UP:
		window.makeDisplacer(0.5, shape.translate, ((0.0, 0.0),), ((0.0, 10.0),))
	elif key == '1':
		# Instead of doing shape.rotate(math.pi / 12.0), we do this.
		window.makeDisplacer(0.5, shape.rotate, (0.0,), (math.pi / 12.0,))
	elif key == '2':
		window.makeDisplacer(0.5, shape.rotate, (0.0,), (-math.pi / 12.0,))
	# The window will automatically redraw itself on each frame of the animation.

def handleClick(xy, name):
	"""If the shift key was down, then sets up a delay to translate the shape that was clicked. If the shift key was not down, then inverts the color of the shape that was clicked."""
	# Use the name to get the Shape object itself.
	window = jr.getWindow()
	shape = window.getBodies2D()[name]
	if window.shiftKeyWasDown():
		print "In two seconds, shape " + str(name) + " will translate."
		pos = shape.getPosition()
		newPos = (pos[0] + 10.0, pos[1] + 10.0)
		# Instead of doing shape.setPosition(newPos), we do this.
		window.makeDelayer(2.0, shape.setPosition, (newPos,))
	else:
		color = shape.getColor()
		newColor = (1.0 - color[0], 1.0 - color[1], 1.0 - color[2], color[3])
		# Instead of doing shape.SetColor(newColor), we do this.
		window.makePositioner(0.5, shape.setColor, (color,), (newColor,))
	# The window will automatically redraw itself on each frame of the animation and when the delay fires.

if __name__ == "__main__":
	# All Bopagopa programs begin by creating a window.
	window = jr.WindowJr()
	# Register the three event handlers defined above.
	window.setKeyboardHandler(handleKeyboard)
	window.setClickHandler2D(handleClick)
	# Make a red disk, scaled by 1.3 in the horizontal direction.
	disk = jr.makeDisk(32.0)
	disk.setColor((1.0, 0.0, 0.0, 0.9))
	disk.setPosition((448.0, 256.0))
	disk.setScales((1.3, 1.0))
	window.appendBody2D(disk)
	# Make a yellow sector.
	sector = jr.makeSector(48.0, 0.0, (2.0 / 3.0) * math.pi)
	sector.setColor((1.0, 1.0, 0.0, 0.9))
	sector.setPosition((352.0, 422.0))
	window.appendBody2D(sector)
	# Make an green annulus.
	annulus = jr.makeAnnulus(24.0, 48.0)
	annulus.setColor((0.0, 1.0, 0.0, 0.9))
	annulus.setPosition((160.0, 422.0))
	window.appendBody2D(annulus)
	# Make a cyan rectangle, rotated pi / 6 = 30 degrees counterclockwise.
	rectangle = jr.makeRectangle(64.0, 32.0)
	rectangle.setColor((0.0, 1.0, 1.0, 0.9))
	rectangle.setPosition((64.0, 256.0))
	rectangle.rotate(math.pi / 6.0)
	window.appendBody2D(rectangle)
	# Make a blue rounded rectangle.
	roundedRectangle = jr.makeRoundedRectangle(64.0, 96.0, 8.0)
	roundedRectangle.setColor((0.0, 0.0, 1.0, 0.9))
	roundedRectangle.setPosition((160.0, 90.0))
	window.appendBody2D(roundedRectangle)
	# Make a magenta polygon.
	vertices = ((41.0, 5.0), (20.0, 50.0), (-70.0, 10.0), (-10.0, -60.0), (30.0, -30.0))
	polygon = jr.makeConvexPolygon(vertices)
	polygon.setColor((1.0, 0.0, 1.0, 0.9))
	polygon.setPosition((352.0, 90.0))
	window.appendBody2D(polygon)
	# All Bopagopa programs end by running the user interface.
	jr.beginUserInterface()
