# 2010 July 7 by Joshua R. Davis

"""
This program demonstrates how to make a simple simulation or game that operates in real time. The graphics are identical to those in the preceding tutorial, but now they belong to a simulation object, which hooks into Bopagopa's animation mechanism (to track the passage of time) and Bopagopa's keyboard handling mechanism (to let the user control the ship using the arrow keys).
"""

import bopagopa.bopagopajr as jr
import math
import random
import OpenGL.GLUT as glut

class Simulation(object):
	"""There is one Simulation object per simulation. The Simulation object possesses all of the other objects (spaceship, asteroids), updates their positions based on the passage of time, and responds to user interface events."""
	
	def __init__(self):
		"""Creates the simulation object."""
		# Construct the spaceship, as in the previous tutorial.
		plume = jr.makeConvexPolygon([(0.0, 0.0), (0.0, 32.0), (-32.0, 16.0)])
		plume.setColor((0.8, 0.8, 1.0, 0.5))
		plume.setPosition((-32.0, -16.0))
		rectangle = jr.makeRectangle(16.0, 32.0)
		rectangle.setColor((0.8, 0.8, 0.8, 1.0))
		rectangle.setPosition((-24.0, -0.0))
		left = jr.makeConvexPolygon([(0.0, 0.0), (23.0, 0.0), (0.0, 23.0)])
		left.setColor((0.8, 0.8, 0.8, 1.0))
		left.setPosition((0.0, 23.0))
		right = jr.makeConvexPolygon([(0.0, 0.0), (0.0, -23.0), (23.0, 0.0)])
		right.setColor((0.8, 0.8, 0.8, 1.0))
		right.setPosition((0.0, -23.0))
		disk = jr.makeDisk(32.0)
		disk.setColor((0.8, 0.8, 0.8, 1.0))
		sector = jr.makeSector(24.0, -math.pi / 4.0, math.pi / 4.0)
		sector.setColor((0.0, 0.0, 0.1, 1.0))
		group = jr.Group2D([plume, rectangle, left, right, disk, sector])
		group.setPosition((256.0, 256.0))
		window.appendBody2D(group)
		# On second thought, make the exhaust plume transparent.
		plume.setColor((0.8, 0.8, 1.0, 0.0))
		# Store the ship's velocity. (No, storing it outside the ship is not great object-oriented design.)
		self.velocities = [[0.0, 0.0]]
		# Make some asteroids with random velocities.
		moonTexture = jr.makeFileTexture("256moon.jpg")
		for i in range(5):
			asteroid = jr.makeDisk(random.random() * 40.0 + 10.0)
			asteroid.setPosition((random.random() * 512.0, random.random() * 512.0))
			asteroid.setColor((0.5, 0.5, 0.5, 1.0))
			asteroid.setTexture(moonTexture)
			window.appendBody2D(asteroid)
			self.velocities.append([random.random() * 20.0 - 10.0, random.random() * 20.0 - 10.0])
		# Register with the window to track the passage of time.
		window.trackTime(self.step)
	
	def handleKeyboard(self, xy, key):
		"""Steers the ship."""
		ship = jr.getWindow().getBodies2D()[0]
		if key == glut.GLUT_KEY_LEFT:
			ship.rotate(math.pi / 24.0)
		elif key == glut.GLUT_KEY_RIGHT:
			ship.rotate(-math.pi / 24.0)
		elif key == glut.GLUT_KEY_UP:
			# Add some velocity in the direction the ship is facing.
			angle = ship.getAngle()
			self.velocities[0][0] += 5.0 * math.cos(angle)
			self.velocities[0][1] += 5.0 * math.sin(angle)
			# Show the exhaust plume.
			plume = ship.getBodies()[0]
			rgba = plume.getColor()
			plume.setColor((rgba[0], rgba[1], rgba[2], 1.0))
		# Inform the window that it should redraw itself, because the graphics have been changed.
		jr.getWindow().needsRedrawing()
	
	def step(self, delta):
		"""The delta argument is the amount of time, in seconds, that has passed since the previous frame of the simulation. Translates each shape by delta times its velocity vector."""
		# Compute the time passed since the last step.
		window = jr.getWindow()
		bodies = window.getBodies2D()
		# Increment position by delta times velocity.
		for i in range(len(self.velocities)):
			bodies[i].translate((delta * self.velocities[i][0], delta * self.velocities[i][1]))
		# Fade the exhaust plume over 1/2 second.
		plume = bodies[0].getBodies()[0]
		rgba = plume.getColor()
		plume.setColor((rgba[0], rgba[1], rgba[2], max(0.0, rgba[3] - delta * 2.0)))
		# The animation system automatically notifies the window that it needs to redraw itself.

if __name__ == "__main__":
	# All Bopagopa programs begin by creating a window. This time, turn off shape dragging.
	window = jr.WindowJr()
	window.setDragHandler2D(None)
	# Create a game object; tell it to handle the keyboard too.
	simulation = Simulation()
	window.setKeyboardHandler(simulation.handleKeyboard)
	# All Bopagopa programs end by running the user interface.
	jr.beginUserInterface()
