# 2010 September 10 by Joshua R. Davis

"""
Like the preceding tutorial, this program demonstrates how to make a simple simulation or game. However, there are three major changes. First, the spaceship and asteroids bounce off the edges of the window. Second, the user controls the spaceship using the mouse rather than the keyboard. Whenever the mouse button is down, the spaceship accelerates toward the mouse pointer. Third, the spaceship is hand-drawn and rendered as a single textured rectangle. (In classic video-game parlance, it is a "sprite".) This allows for a much more detailed spaceship. In computer graphics, adding detail through textures typically leads to higher performance than adding detail through triangles.
"""

import bopagopa.bopagopajr as jr
import math
import random
import OpenGL.GLUT as glut

class Simulation(object):
	"""There is one Simulation object per simulation. The Simulation object possesses all of the other objects (spaceship, asteroids), updates their positions based on the passage of time, and responds to user interface events."""
	
	def __init__(self):
		"""Creates the simulation object."""
		# Construct the spaceship as a single textured rectangle, a.k.a. sprite.
		sprite = jr.makeRectangle(64.0, 64.0)
		sprite.setTexture(jr.makeFileTexture("64spaceship.png"))
		# Construct the plume.
		plume = jr.makeConvexPolygon([(0.0, 0.0), (0.0, 16.0), (-32.0, 8.0)])
		plume.setPosition((-25.0, -8.0))
		plume.setColor((0.8, 0.8, 1.0, 0.0))
		# Join the ship sprite and plume.
		spaceship = jr.Group2D([plume, sprite])
		spaceship.setPosition((256.0, 256.0))
		window.appendBody2D(spaceship)
		# Store the ship's velocity. (No, storing it outside the ship is not great object-oriented design.)
		self.velocities = [[0.0, 0.0]]
		# Make some asteroids with random velocities.
		moonTexture = jr.makeFileTexture("256moon.jpg")
		for i in range(5):
			asteroid = jr.makeDisk(random.random() * 40.0 + 10.0)
			asteroid.setPosition((random.random() * 512.0, random.random() * 512.0))
			asteroid.setColor((0.5, 0.5, 0.5, 1.0))
			asteroid.setTexture(moonTexture)
			window.appendBody2D(asteroid)
			self.velocities.append([random.random() * 20.0 - 10.0, random.random() * 20.0 - 10.0])
		# Register with the window to track the passage of time.
		window.trackTime(self.step)
	
	def step(self, delta):
		"""The delta argument is the amount of time, in seconds, that has passed since the previous frame of the simulation. Translates each shape by delta times its velocity vector. Bounces the shapes off walls. Accelerates the ship toward the mouse pointer, if the mouse button is down."""
		window = jr.getWindow()
		bodies = window.getBodies2D()
		plume = bodies[0].getBodies()[0]
		rgba = plume.getColor()
		# Turn the spaceship toward the mouse pointer.
		xy = bodies[0].getPosition()
		xyMouse = window.getMouse()
		xyDelta = (xyMouse[0] - xy[0], xyMouse[1] - xy[1])
		if xyDelta != (0, 0):
			angle = math.atan2(xyDelta[1], xyDelta[0])
			bodies[0].setAngle(angle)
		if window.mouseIsDown():
			# Accelerate the spaceship toward the mouse pointer, if the button is down.
			self.velocities[0] = [self.velocities[0][0] + delta * xyDelta[0], self.velocities[0][1] + delta * xyDelta[1]]
			plume.setColor((rgba[0], rgba[1], rgba[2], 1.0))
		else:
			# Fade the exhaust plume over 1/2 second.
			plume.setColor((rgba[0], rgba[1], rgba[2], max(0.0, rgba[3] - delta * 2.0)))
		for i in range(len(self.velocities)):
			# Increment position by delta times velocity.
			xy = bodies[i].getPosition()
			xy = [xy[0] + delta * self.velocities[i][0], xy[1] + delta * self.velocities[i][1]]
			# Bounce off walls.
			if xy[0] < 0.0:
				xy[0] = 0.0
				self.velocities[i][0] = -1 * self.velocities[i][0]
			elif xy[0] > window.getWidth():
				xy[0] = window.getWidth()
				self.velocities[i][0] = -1 * self.velocities[i][0]
			if xy[1] < 0:
				xy[1] = 0.0
				self.velocities[i][1] = -1 * self.velocities[i][1]
			elif xy[1] > window.getHeight():
				xy[1] = window.getHeight()
				self.velocities[i][1] = -1 * self.velocities[i][1]
			bodies[i].setPosition(xy)
		# The animation system automatically notifies the window that it needs to redraw itself.

if __name__ == "__main__":
	# All Bopagopa programs begin by creating a window.
	window = jr.WindowJr()
	window.setDragHandler2D(None)
	simulation = Simulation()
	# All Bopagopa programs end by running the user interface.
	jr.beginUserInterface()
