# 2010 October 3 by Joshua R. Davis

"""
This tutorial expands on the preceding tutorial by adding the Earth's moon to the simulation. The moon's center of revolution (the Earth) changes over time, so we can't animate the moon's revolution about the Earth using a simple animation. Instead, we set up a simulation object that tracks time and manually animates all three bodies. As in the previous tutorial, some aspects of the simulation are realistic, and others (especially scale) not. One second of the simulation represents one day in reality.
"""

import math
import bopagopa.bopagopajr as jr

class Simulation(object):
	"""Owns all of the objects and simulates the passage of time on them."""
	
	def __init__(self):
		"""Initializes sun, earth, and moon, and tracks time."""
		window = jr.getWindow()
		# Make the sun and earth.
		sun = jr.makeSphere(1.0)
		axes = jr.makeAxes(2.0, 0.01)
		self.sunGroup = jr.Group3D(axes + (sun,))
		window.appendBody3D(self.sunGroup)
		earth = jr.makeSphere(0.25)
		earthTexture = jr.makeFileTexture("256earth.jpg")
		earth.setTexture(earthTexture)
		axes = jr.makeAxes(0.5, 0.01)
		self.earthGroup = jr.Group3D(axes + (earth,))
		window.appendBody3D(self.earthGroup)
		moon = jr.makeSphere(0.0625)
		moonTexture = jr.makeFileTexture("256moon.jpg")
		moon.setTexture(moonTexture)
		axes = jr.makeAxes(0.125, 0.01)
		self.moonGroup = jr.Group3D(axes + (moon,))
		window.appendBody3D(self.moonGroup)
		# Set their initial positions and orientations.
		self.earthGroup.setPosition((2.0, 0.0, 0.0))
		self.earthGroup.rotate(23.44 * math.pi / 180.0, (1.0, 0.0, 0.0))
		self.moonGroup.setPosition((2.5, 0.0, 0.0))
		self.moonGroup.revolve(5.14 * math.pi / 180.0, (0.0, 1.0, 0.0), self.earthGroup.getPosition())
		self.moonGroup.rotate(6.18 * math.pi / 180.0, (0.0, 1.0, 0.0))
		# Store the rotation axes for later use.
		self.earthRotationAxis = self.earthGroup.vectorToWorld((0.0, 0.0, 1.0))
		self.moonRevolutionAxis = (math.sin(5.14 * math.pi / 180.0), 0.0, math.cos(5.14 * math.pi / 180.0))
		self.moonRotationAxis = (math.sin(6.18 * math.pi / 180.0), 0.0, math.cos(6.18 * math.pi / 180.0))
		# Create the animation that keeps track of time.
		window.trackTime(self.step)
	
	def step(self, delta):
		"""Performs one step of simulation and animation."""
		angleChange = delta * 2.0 * math.pi
		# Revolve the earth about the sun once every 365.0 seconds.
		oldEarthPos = self.earthGroup.getPosition()
		self.earthGroup.revolve(angleChange / 365.0, (0.0, 0.0, 1.0), (0.0, 0.0, 0.0))
		newEarthPos = self.earthGroup.getPosition()
		translation = jr.vectorDifference(newEarthPos, oldEarthPos)
		# Rotate the earth about its axis once every second.
		self.earthGroup.rotate(angleChange, self.earthRotationAxis)
		# Revolve the moon about the earth once every 27.25 seconds.
		self.moonGroup.translate(translation)
		self.moonGroup.revolve(angleChange / 27.25, self.moonRevolutionAxis, self.earthGroup.getPosition())
		# Rotate the moon on its axis at the same rate.
		self.moonGroup.rotate(angleChange / 27.25, self.moonRotationAxis)

if __name__ == "__main__":
	# All Bopagopa programs begin by creating a window. We also register our user interface handlers here.
	window = jr.WindowJr()
	window.setBeginDragHandler3D(None)
	window.setDragHandler3D(None)
	# Make a light at the origin; also increase the ambient light to 20% intensity.
	light = jr.Light()
	light.setPosition((0.0, 0.0, 0.0))
	window.setLights((light,))
	window.setAmbientColor((0.2, 0.2, 0.2))
	# Create the simulation object, which does all of the work.
	simulation = Simulation()
	# All Bopagopa programs end by running the user interface.
	jr.beginUserInterface()
