# 2010 October 3 by Joshua R. Davis

"""
Bopagopa comes with three kinds of lights: Light (a subclass of Camera), DirectionaLight (a subclass of Light), and SpotLight (a subclass of DirectionalLight). Light represents an omnidirectional point light source positioned somewhere in the scene; its position matters to the lighting of the scene, but its orientation does not, because it casts light in all directions. DirectionalLight represents a distant, powerful light source (such as the sun or moon) whose rays arrive on the scene parallel to each other; its orientation matters, but its position does not. SpotLight represents a light positioned somewhere in the scene that casts a cone of light; its position and orientation both matter. Light does not support shadow casting, but DirectionalLight and SpotLight do.

This tutorial demonstrates DirectionalLight and SpotLight. The menu lets the user control a few aspects such as color and spot light angle. On a surface where both lights shine, the color range gets saturated; this could be fixed by better choice of surface or light color. Also, the directional light can shadow only those objects within its shadow-casting volume; objects outside this volume are always lit. Currently the shadow-casting volume must be set manually using Camera.containPoints. In this program, containPoints is used once, to make the initial shadow-casting volume contain the entire ground box. So shadow effects are correct at the start of the program, but if you move some objects or move the light, then some shadow effects may go wrong, because the shadow-casting volume is not being recomputed.
"""

import bopagopa.bopagopajr as jr
import math

def handleMenu(itemName):
	"""Responds to the menu that is tied to the right mouse button."""
	window = jr.getWindow()
	if itemName == "Directional off":
		light = window.getLights()[0]
		window.makePositioner(0.5, light.setColor, (light.getColor(),), ((0.0, 0.0, 0.0),))
	elif itemName == "Directional white":
		light = window.getLights()[0]
		window.makePositioner(0.5, light.setColor, (light.getColor(),), ((1.0, 1.0, 1.0),))
	elif itemName == "Directional red":
		light = window.getLights()[0]
		window.makePositioner(0.5, light.setColor, (light.getColor(),), ((1.0, 0.0, 0.0),))
	elif itemName == "Directional revolve":
		light = window.getLights()[0]
		window.makeDisplacer(0.5, light.revolveRotate, (0.0, (0.0, 0.0, 1.0), (0.0, 0.0, 0.0)), (math.pi / 12.0, (0.0, 0.0, 0.0), (0.0, 0.0, 0.0)))
	elif itemName == "Spot off":
		light = window.getLights()[1]
		window.makePositioner(0.5, light.setColor, (light.getColor(),), ((0.0, 0.0, 0.0),))
	elif itemName == "Spot white":
		light = window.getLights()[1]
		window.makePositioner(0.5, light.setColor, (light.getColor(),), ((1.0, 1.0, 1.0),))
	elif itemName == "Spot narrow":
		light = window.getLights()[1]
		window.makePositioner(0.5, light.setAngle, (light.getAngle(),), (math.pi / 12.0,))
	elif itemName == "Spot wide":
		light = window.getLights()[1]
		window.makePositioner(0.5, light.setAngle, (light.getAngle(),), (math.pi / 4.0,))

if __name__ == "__main__":
	# All Bopagopa programs begin by creating a window.
	window = jr.WindowJr()
	window.setMenuHandler(handleMenu, ["Directional off", "Directional white", "Directional red", "Directional revolve", "Spot off", "Spot white", "Spot narrow", "Spot wide"])
	# Make some lights with shadow maps.
	directional = jr.DirectionalLight(1024)
	directional.revolveRotate(math.pi / 4.0, (0.0, 1.0, 0.0), (0.0, 0.0, 0.0))
	directional.containPoints([(-5.0, -5.0, 0.0), (5.0, -5.0, 0.0), (-5.0, 5.0, 0.0), (5.0, 5.0, 0.0), (-5.0, -5.0, -1.0), (5.0, -5.0, -1.0), (-5.0, 5.0, -1.0), (5.0, 5.0, -1.0)])
	spot = jr.SpotLight(1024)
	spot.revolveRotate(math.pi / 4.0, (1.0, 0.0, 0.0), (0.0, 0.0, 0.0))
	window.setLights((directional, spot))
	# Make some solids.
	box = jr.makeBox(10.0, 10.0, 1.0)
	box.setColor((0.0, 1.0, 0.0))
	box.setPosition((0.0, 0.0, -1.0))
	window.appendBody3D(box)
	sphere = jr.makeSphere(1.0)
	sphere.setColor((1.0, 1.0, 0.0))
	window.appendBody3D(sphere)
	cylinder = jr.makeCylinder(1.0, 0.5)
	cylinder.setColor((0.0, 0.0, 1.0))
	cylinder.setPosition((1.0, 0.0, 1.0))
	window.appendBody3D(cylinder)
	capsule = jr.makeCapsule(1.0, 0.5)
	capsule.setColor((1.0, 0.0, 1.0))
	capsule.setPosition((0.0, 1.0, 1.0))
	window.appendBody3D(capsule)
	cone = jr.makeCone(1.0, 0.5)
	cone.setColor((1.0, 0.5, 0.0))
	cone.setPosition((1.0, 1.0, 1.0))
	window.appendBody3D(cone)
	# All Bopagopa programs end by running the user interface.
	jr.beginUserInterface()
