# 2010 October 3 by Joshua R. Davis

"""
Bopagopa itself doesn't provide any dynamics or collision detection. However, Bopagopa is loosely integrated with the Open Dynamics Engine (ODE), in that Bopagopa's 3D bodies use ODE bodies to keep track of their positions and orientations. These bodies exist in an ODE world that is owned by the window. Bopagopa lets you access the window's world and the bodies' bodies; with these you can use all of ODE.

This tutorial shows a compound pendulum made of two bodies that are joined but do not collide with each other. If you leave the pendulum alone, then it acts as a planar pendulum (with hinge joints). If you perturb the bodies a little with the mouse, then they may leave the plane; then it becomes clear that the joints are balls and sockets, not hinges. Gravitational acceleration should be -9.81 m / s^2, but I've weakened it by a factor of 10; full gravity leads to instability for some reason. (Possibly we should be using shorter time steps for the dynamical simulation --- say, 10 simulation steps for every animation frame. I haven't worked with ODE much.)
"""

import bopagopa.bopagopajr as jr
import ode
import math

if __name__ == "__main__":
	# All Bopagopa programs begin by creating a window.
	window = jr.WindowJr()
	# Set up the window to have gravity, albeit weak gravity.
	window.getWorld().setGravity((0.0, 0.0, -9.81 / 10.0))
	# Tell the dynamical simulation to run, in lockstep with the window animation.
	window.trackTime(window.getWorld().step)
	# Make lights and coordinate axes.
	light = jr.Light()
	window.setLights((light,))
	axes = jr.makeAxes(10.0, 0.1)
	frame = jr.Group3D(axes)
	window.appendBody3D(frame)
	# Make the camera, lights, and coordinate frame immune to gravity.
	window.getCamera().setGravityMode(False)
	for light in window.getLights():
		light.setGravityMode(False)
	for axis in axes:
		axis.setGravityMode(False)
	frame.setGravityMode(False)
	# Make a capsule.
	capsule = jr.makeCapsule(1.0, 0.5)
	capsule.setColor((1.0, 0.0, 1.0))
	capsule.rotate(-math.pi / 2.0, (0.0, 1.0, 0.0))
	capsule.setPosition((0.5, 0.0, 0.0))
	window.appendBody3D(capsule)
	# Give the capsule mass.
	capsuleMass = ode.Mass()
	capsuleMass.setCappedCylinder(2500.0, 3, 0.5, 1.0)
	capsule.setMass(capsuleMass)
	# Fix one end of the capsule to the world.
	capsuleJoint = ode.BallJoint(window.getWorld())
	capsuleJoint.attach(capsule.getBody(), None)
	capsuleJoint.setAnchor((0.0, 0.0, 0.0))
	#capsule.addForce((0.0, 200.0, 0.0))
	# Make a cylinder.
	cylinder = jr.makeCylinder(1.0, 0.5)
	cylinder.setColor((0.0, 0.0, 1.0))
	cylinder.rotate(-math.pi / 2.0, (0.0, 1.0, 0.0))
	cylinder.setPosition((1.5, 0.0, 0.0))
	window.appendBody3D(cylinder)
	# Give the cylinder mass.
	cylinderMass = ode.Mass()
	cylinderMass.setCylinder(2500.0, 3, 0.5, 1.0)
	cylinder.setMass(cylinderMass)
	# Fix one end of the cylinder to the other end of the capsule.
	cylinderJoint = ode.BallJoint(window.getWorld())
	cylinderJoint.attach(capsule.getBody(), cylinder.getBody())
	cylinderJoint.setAnchor((1.0, 0.0, 0.0))
	# All Bopagopa programs end by telling the window to run the user interface.
	jr.beginUserInterface()
